# Generated with SMOP  0.41
from libsmop import *
# solar_ET.m

    
@function
def solar_ET(d=None,*args,**kwargs):
    varargin = solar_ET.varargin
    nargin = solar_ET.nargin

    # used to calculate the equation of time for use in calcuating the zenith
# angle of the sun.
    
    f=279.575 + multiply(0.9856,d)
# solar_ET.m:7
    ET=(multiply(- 104.7,sind(f)) + multiply(596.2,sind(dot(2.0,f))) + multiply(4.3,sind(dot(3.0,f))) - multiply(12.7,sind(dot(4.0,f))) - multiply(429.3,cosd(f)) - multiply(2.0,cosd(dot(2.0,f))) + multiply(19.3,cosd(dot(3.0,f)))) / 3600
# solar_ET.m:9


from __future__ import absolute_import
import numpy as np
from .Kvikt_veg import Kvikt_veg
from . import sunlit_shaded_patches

def Kside_veg_v2022a(radI,radD,radG,
                    shadow,svfS,svfW,svfN,svfE,svfEveg,svfSveg,svfWveg,svfNveg,
                    azimuth,altitude,psi,t,albedo,F_sh,
                    KupE,KupS,KupW,KupN,
                    cyl,lv,anisotropic_diffuse,diffsh,rows,cols,asvf,
                    shmat, vegshmat, vbshvegshmat):

    # New reflection equation 2012-05-25
    vikttot=4.4897
    aziE=azimuth+t
    aziS=azimuth-90+t
    aziW=azimuth-180+t
    aziN=azimuth-270+t
    deg2rad=np.pi/180
    KsideD = np.zeros((rows, cols))
    Kref_sun = np.zeros((rows, cols))
    Kref_sh = np.zeros((rows, cols))
    Kref_veg = np.zeros((rows, cols))
    Kside = np.zeros((rows, cols))

    Kref_veg_n = np.zeros((rows, cols))
    Kref_veg_s = np.zeros((rows, cols))
    Kref_veg_e = np.zeros((rows, cols))
    Kref_veg_w = np.zeros((rows, cols))

    Kref_sh_n = np.zeros((rows, cols))
    Kref_sh_s = np.zeros((rows, cols))
    Kref_sh_e = np.zeros((rows, cols))
    Kref_sh_w = np.zeros((rows, cols))

    Kref_sun_n = np.zeros((rows, cols))
    Kref_sun_s = np.zeros((rows, cols))
    Kref_sun_e = np.zeros((rows, cols))
    Kref_sun_w = np.zeros((rows, cols))

    KeastRef = np.zeros((rows, cols)); KwestRef = np.zeros((rows, cols)); KnorthRef = np.zeros((rows, cols)); KsouthRef = np.zeros((rows, cols))
    diffRadE = np.zeros((rows, cols)); diffRadS = np.zeros((rows, cols)); diffRadW = np.zeros((rows, cols)); diffRadN = np.zeros((rows, cols))

    ### Direct radiation ###
    if cyl == 1: ### Kside with cylinder ###
        KsideI=shadow*radI*np.cos(altitude*deg2rad)
        KeastI=0;KsouthI=0;KwestI=0;KnorthI=0
    else: ### Kside with weights ###
        if azimuth > (360-t)  or  azimuth <= (180-t):
            KeastI=radI*shadow*np.cos(altitude*deg2rad)*np.sin(aziE*deg2rad)
        else:
            KeastI=0
        if azimuth > (90-t)  and  azimuth <= (270-t):
            KsouthI=radI*shadow*np.cos(altitude*deg2rad)*np.sin(aziS*deg2rad)
        else:
            KsouthI=0
        if azimuth > (180-t)  and  azimuth <= (360-t):
            KwestI=radI*shadow*np.cos(altitude*deg2rad)*np.sin(aziW*deg2rad)
        else:
            KwestI=0
        if azimuth <= (90-t)  or  azimuth > (270-t):
            KnorthI=radI*shadow*np.cos(altitude*deg2rad)*np.sin(aziN*deg2rad)
        else:
            KnorthI=0

        KsideI=shadow*0

    ### Diffuse and reflected radiation ###
    [viktveg,viktwall]=Kvikt_veg(svfE,svfEveg,vikttot)
    svfviktbuvegE=(viktwall+(viktveg)*(1-psi))

    [viktveg,viktwall]=Kvikt_veg(svfS,svfSveg,vikttot)
    svfviktbuvegS=(viktwall+(viktveg)*(1-psi))

    [viktveg,viktwall]=Kvikt_veg(svfW,svfWveg,vikttot)
    svfviktbuvegW=(viktwall+(viktveg)*(1-psi))

    [viktveg,viktwall]=Kvikt_veg(svfN,svfNveg,vikttot)
    svfviktbuvegN=(viktwall+(viktveg)*(1-psi))

    ### Anisotropic Diffuse Radiation after Perez et al. 1993 ###
    if anisotropic_diffuse == 1:

        anisotropic_sky = True

        patch_altitude = lv[:, 0]
        patch_azimuth = lv[:, 1]
        if anisotropic_sky:
            patch_luminance = lv[:, 2]
        else:
            patch_luminance = np.zeros((patch_altitude.shape[0]))
            patch_luminance[:] = 1.0/patch_luminance.shape[0]

        # Unique altitudes for patches
        skyalt, skyalt_c = np.unique(patch_altitude, return_counts=True)

        radTot = np.zeros(1)

        # Calculation of steradian for each patch
        steradian = np.zeros((patch_altitude.shape[0]))
        for i in range(patch_altitude.shape[0]):
            # If there are more than one patch in a band
            if skyalt_c[skyalt == patch_altitude[i]] > 1:
                steradian[i] = ((360 / skyalt_c[skyalt == patch_altitude[i]]) * deg2rad) * (np.sin((patch_altitude[i] + patch_altitude[0]) * deg2rad) \
                - np.sin((patch_altitude[i] - patch_altitude[0]) * deg2rad))
            # If there is only one patch in band, i.e. 90 degrees
            else:
                steradian[i] = ((360 / skyalt_c[skyalt == patch_altitude[i]]) * deg2rad) * (np.sin((patch_altitude[i]) * deg2rad) \
                    - np.sin((patch_altitude[i-1] + patch_altitude[0]) * deg2rad))

            radTot += (patch_luminance[i] * steradian[i] * np.sin(patch_altitude[i] * deg2rad)) # Radiance fraction normalization

        lumChi = (patch_luminance * radD) / radTot # Radiance fraction normalization

        if cyl == 1:
            for idx in range(patch_azimuth.shape[0]):
                # Angle of incidence, np.cos(0) because cylinder - always perpendicular
                anglIncC = np.cos(patch_altitude[idx] * deg2rad) * np.cos(0) # * np.sin(np.pi / 2) \
                    # + np.sin(patch_altitude[idx] * deg2rad) * np.cos(np.pi / 2)
                # Diffuse vertical radiation
                KsideD += diffsh[:, :, idx] * lumChi[idx] * anglIncC * steradian[idx]

                # Shortwave reflected on sunlit surfaces
                # sunlit_surface = ((albedo * radG) / np.pi)
                sunlit_surface = ((albedo * (radI * np.cos(altitude * deg2rad)) + (radD * 0.5)) / np.pi)
                # Shortwave reflected on shaded surfaces and vegetation
                shaded_surface = ((albedo * radD * 0.5) / np.pi)
                
                # Shortwave radiation reflected on vegetation - based on diffuse shortwave radiation
                temp_vegsh = ((vegshmat[:,:,idx] == 0) | (vbshvegshmat[:,:,idx] == 0))
                Kref_veg += shaded_surface * temp_vegsh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)

                # Shortwave radiation reflected on buildings (shaded and sunlit) - based on global and diffuse shortwave radiation
                temp_vbsh = (1 - shmat[:,:,idx]) * vbshvegshmat[:,:,idx]
                temp_sh = (temp_vbsh == 1) # & (vbshvegshmat[:,:,idx] == 1)

                sunlit_patches, shaded_patches = sunlit_shaded_patches.shaded_or_sunlit(altitude, azimuth, patch_altitude[idx], patch_azimuth[idx], asvf)
                Kref_sun += sunlit_surface * sunlit_patches * temp_sh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)
                Kref_sh += shaded_surface * shaded_patches * temp_sh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)

            Kside = KsideI + KsideD + Kref_sun + Kref_sh + Kref_veg

            Keast = (KupE * 0.5)
            Kwest = (KupW * 0.5)
            Knorth = (KupN * 0.5)
            Ksouth = (KupS * 0.5)

            # Keast  = (albedo * (svfviktbuvegE * (radG * (1 - F_sh) + radD * F_sh)) + KupE) * 0.5
            # Ksouth = (albedo * (svfviktbuvegS * (radG * (1 - F_sh) + radD * F_sh)) + KupS) * 0.5
            # Kwest  = (albedo * (svfviktbuvegW * (radG * (1 - F_sh) + radD * F_sh)) + KupW) * 0.5
            # Knorth = (albedo * (svfviktbuvegN * (radG * (1 - F_sh) + radD * F_sh)) + KupN) * 0.5
        else: # Box
            diffRadE = np.zeros((rows, cols)); diffRadS = np.zeros((rows, cols)); diffRadW = np.zeros((rows, cols)); diffRadN = np.zeros((rows, cols))

            for idx in range(patch_azimuth.shape[0]):
                if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] <= 180):
                    anglIncE = np.cos(patch_altitude[idx] * deg2rad) * np.cos((90 - patch_azimuth[idx] + t) * deg2rad) # * np.sin(np.pi / 2) \
                        # + np.sin(patch_altitude[idx] * deg2rad) * np.cos(np.pi / 2)
                    diffRadE += diffsh[:, :, idx] * lumChi[idx] * anglIncE * steradian[idx] #* 0.5

                if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] <= 270):
                    anglIncS = np.cos(patch_altitude[idx] * deg2rad) * np.cos((180 - patch_azimuth[idx] + t) * deg2rad) # * np.sin(np.pi / 2) \
                        # + np.sin(patch_altitude[idx] * deg2rad) * np.cos(np.pi / 2)
                    diffRadS += diffsh[:, :, idx] * lumChi[idx] * anglIncS * steradian[idx] #* 0.5

                if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] <= 360):
                    anglIncW = np.cos(patch_altitude[idx] * deg2rad) * np.cos((270 - patch_azimuth[idx] + t) * deg2rad) # * np.sin(np.pi / 2) \
                        # + np.sin(patch_altitude[idx] * deg2rad) * np.cos(np.pi / 2)
                    diffRadW += diffsh[:, :, idx] * lumChi[idx] * anglIncW * steradian[idx] #* 0.5

                if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] <= 90):
                    anglIncN = np.cos(patch_altitude[idx] * deg2rad) * np.cos((0 - patch_azimuth[idx] + t) * deg2rad) # * np.sin(np.pi / 2) \
                        # + np.sin(patch_altitude[idx] * deg2rad) * np.cos(np.pi / 2)
                    diffRadN += diffsh[:, :, idx] * lumChi[idx] * anglIncN * steradian[idx] #* 0.5
            
                # Shortwave reflected on sunlit surfaces
                # sunlit_surface = ((albedo * radG) / np.pi)
                sunlit_surface = ((albedo * (radI * np.cos(altitude * deg2rad)) + (radD * 0.5)) / np.pi)
                # Shortwave reflected on shaded surfaces and vegetation
                shaded_surface = ((albedo * radD * 0.5) / np.pi)
                
                # Shortwave radiation reflected on vegetation - based on diffuse shortwave radiation
                temp_vegsh = ((vegshmat[:,:,idx] == 0) | (vbshvegshmat[:,:,idx] == 0))
                Kref_veg += shaded_surface * temp_vegsh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)

                if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
                    Kref_veg_e += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((90 - patch_azimuth[idx] + t) * deg2rad)
                if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
                    Kref_veg_s += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((180 - patch_azimuth[idx] + t) * deg2rad)
                if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
                    Kref_veg_w += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((270 - patch_azimuth[idx] + t) * deg2rad)
                if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
                    Kref_veg_n += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((0 - patch_azimuth[idx] + t) * deg2rad)

                # Shortwave radiation reflected on buildings (shaded and sunlit) - based on global and diffuse shortwave radiation
                temp_vbsh = (1 - shmat[:,:,idx]) * vbshvegshmat[:,:,idx]
                temp_sh = (temp_vbsh == 1) # & (vbshvegshmat[:,:,idx] == 1)
                azimuth_difference = np.abs(azimuth - patch_azimuth[idx])

                if ((azimuth_difference > 90) and (azimuth_difference < 270)):
                    sunlit_patches, shaded_patches = sunlit_shaded_patches.shaded_or_sunlit(altitude, azimuth, patch_altitude[idx], patch_azimuth[idx], asvf)
                    Kref_sun += sunlit_surface * sunlit_patches * temp_sh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)
                    Kref_sh += shaded_surface * shaded_patches * temp_sh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)

                    if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
                        Kref_sun_e += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx] + t) * deg2rad)
                        Kref_sh_e += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx] + t) * deg2rad)
                    if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
                        Kref_sun_s += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx] + t) * deg2rad)
                        Kref_sh_s += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx] + t) * deg2rad)
                    if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
                        Kref_sun_w += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx] + t) * deg2rad)
                        Kref_sh_w += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx] + t) * deg2rad)
                    if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
                        Kref_sun_n += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx] + t) * deg2rad)
                        Kref_sh_n += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx] + t) * deg2rad)
                else:
                    Kref_sh += shaded_surface * temp_sh * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad)

                    if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
                        Kref_sh_e += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx] + t) * deg2rad)
                    if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
                        Kref_sh_s += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx] + t) * deg2rad)
                    if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
                        Kref_sh_w += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx] + t) * deg2rad)
                    if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
                        Kref_sh_n += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx] + t) * deg2rad)

            Keast = KeastI + diffRadE + Kref_sun_e + Kref_sh_e + Kref_veg_e + KupE * 0.5
            Kwest = KwestI + diffRadW + Kref_sun_w + Kref_sh_w + Kref_veg_w + KupW * 0.5
            Knorth = KnorthI + diffRadN + Kref_sun_n + Kref_sh_n + Kref_veg_n + KupN * 0.5
            Ksouth = KsouthI + diffRadS + Kref_sun_s + Kref_sh_s + Kref_veg_s + KupS * 0.5

    else:
        KeastDG = (radD * (1 - svfviktbuvegE) + albedo * (
        svfviktbuvegE * (radG * (1 - F_sh) + radD * F_sh)) + KupE) * 0.5
        Keast = KeastI + KeastDG

        KsouthDG = (radD * (1 - svfviktbuvegS) + albedo * (
        svfviktbuvegS * (radG * (1 - F_sh) + radD * F_sh)) + KupS) * 0.5
        Ksouth = KsouthI + KsouthDG

        KwestDG = (radD * (1 - svfviktbuvegW) + albedo * (
        svfviktbuvegW * (radG * (1 - F_sh) + radD * F_sh)) + KupW) * 0.5
        Kwest = KwestI + KwestDG

        KnorthDG = (radD * (1 - svfviktbuvegN) + albedo * (
        svfviktbuvegN * (radG * (1 - F_sh) + radD * F_sh)) + KupN) * 0.5
        Knorth = KnorthI + KnorthDG

    return Keast, Ksouth, Kwest, Knorth, KsideI, KsideD, Kside
    # return Keast, Ksouth, Kwest, Knorth, KsideI, KsideD, Kref_sh, Kref_sun, Kref_veg, Kside


import json
import zipfile
from types import SimpleNamespace
from typing import Any

import numpy as np
import pandas as pd

from ... import common
from ...class_configs import EnvironData, SolweigConfig, SvfData
from ...functions import wallalgorithms as wa
from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
from ...util.SEBESOLWEIGCommonFiles.Solweig_v2015_metdata_noload import Solweig_2015a_metdata_noload
from . import PET_calculations
from . import Solweig_2025a_calc_forprocessing as so
from . import UTCI_calculations as utci
from .CirclePlotBar import PolarBarPlot
from .patch_characteristics import hemispheric_image
from .wall_surface_temperature import load_walls
from .wallsAsNetCDF import walls_as_netcdf


def dict_to_namespace(d):
    """Recursively convert dicts to SimpleNamespace."""
    if isinstance(d, dict):
        return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
    elif isinstance(d, list):
        return [dict_to_namespace(i) for i in d]
    else:
        return d


def Tgmaps_v1(lc_grid, params):
    """
    Populates grids with coefficients for Tg wave based on land cover.
    This is a vectorized version that avoids looping over pixels.
    """
    # Sanitize land cover grid
    lc_grid[lc_grid >= 100] = 2

    # Get unique land cover IDs and filter them
    unique_ids = np.unique(lc_grid)
    valid_ids = unique_ids[unique_ids <= 7].astype(int)

    # Initialize output grids by copying the original land cover grid
    TgK = np.copy(lc_grid)
    Tstart = np.copy(lc_grid)
    alb_grid = np.copy(lc_grid)
    emis_grid = np.copy(lc_grid)
    TmaxLST = np.copy(lc_grid)

    # Create mapping dictionaries from land cover ID to parameter values
    id_to_name = {i: getattr(params.Names.Value, str(i)) for i in valid_ids}
    name_to_tstart = {name: getattr(params.Tstart.Value, name) for name in id_to_name.values()}
    name_to_albedo = {name: getattr(params.Albedo.Effective.Value, name) for name in id_to_name.values()}
    name_to_emissivity = {name: getattr(params.Emissivity.Value, name) for name in id_to_name.values()}
    name_to_tmaxlst = {name: getattr(params.TmaxLST.Value, name) for name in id_to_name.values()}
    name_to_tsdeg = {name: getattr(params.Ts_deg.Value, name) for name in id_to_name.values()}

    # Perform replacements for each valid land cover ID
    for i in valid_ids:
        mask = lc_grid == i
        name = id_to_name[i]
        Tstart[mask] = name_to_tstart[name]
        alb_grid[mask] = name_to_albedo[name]
        emis_grid[mask] = name_to_emissivity[name]
        TmaxLST[mask] = name_to_tmaxlst[name]
        TgK[mask] = name_to_tsdeg[name]

    # Get wall-specific parameters
    TgK_wall = getattr(params.Ts_deg.Value, "Walls", None)
    Tstart_wall = getattr(params.Tstart.Value, "Walls", None)
    TmaxLST_wall = getattr(params.TmaxLST.Value, "Walls", None)

    return TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST, TmaxLST_wall


class SolweigRun:
    """Class to run the SOLWEIG algorithm with given configuration."""

    def __init__(self, config: SolweigConfig, params_json_path: str, qgis_env: bool = True):
        """Initialize the SOLWEIG runner with configuration and parameters."""
        self.config = config
        self.config.validate()
        # Progress tracking settings
        self.progress = None
        self.iters_total: int = 0
        self.iters_count: int = 0
        self.qgis_env = qgis_env
        # Initialize POI data
        self.poi_names: list[Any] = []
        self.poi_pixel_xys: np.ndarray | None = None
        self.poi_results = []
        # Initialize WOI data
        self.woi_names: list[Any] = []
        self.woi_pixel_xys: np.ndarray | None = None
        self.woi_results = []
        # Load parameters from JSON file
        params_path = common.check_path(params_json_path)
        try:
            with open(params_path) as f:
                params_dict = json.load(f)
                self.params = dict_to_namespace(params_dict)
        except Exception as e:
            raise RuntimeError(f"Failed to load parameters from {params_json_path}: {e}")

    def prep_progress(self, num: int) -> None:
        """Prepare progress for environment."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def iter_progress(self) -> bool:
        """Iterate progress ."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_epw_weather(self) -> EnvironData:
        """Load weather data from an EPW file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_met_weather(self, header_rows: int = 1, delim: str = " ") -> EnvironData:
        """Load weather data from a MET file."""
        met_data = np.loadtxt(self.config.met_path, skiprows=header_rows, delimiter=delim)
        return EnvironData(
            YYYY=met_data[:, 0],
            DOY=met_data[:, 1],
            hours=met_data[:, 2],
            minu=met_data[:, 3],
            Ta=met_data[:, 11],
            RH=met_data[:, 10],
            radG=met_data[:, 14],
            radD=met_data[:, 21],
            radI=met_data[:, 22],
            P=met_data[:, 12],
            Ws=met_data[:, 9],
        )

    def load_poi_data(self, trf_arr: list[float]) -> None:
        """Load point of interest (POI) data from a file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def save_poi_results(self, trf_arr: list[float], crs_wkt: str) -> None:
        """Save results for points of interest (POIs) to files."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_woi_data(self, trf_arr: list[float]) -> None:
        """Load wall of interest (WOI) data from a file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def save_woi_results(self, trf_arr: list[float], crs_wkt: str) -> None:
        """Save results for walls of interest (WOIs) to files."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def run(self) -> None:
        """Run the SOLWEIG algorithm."""
        # Load DSM
        dsm_arr, dsm_trf_arr, dsm_crs_wkt, dsm_nd_val = common.load_raster(self.config.dsm_path, bbox=None)
        scale = 1 / dsm_trf_arr[1]
        left_x = dsm_trf_arr[0]
        top_y = dsm_trf_arr[3]
        lng, lat = common.xy_to_lnglat(dsm_crs_wkt, left_x, top_y)
        rows = dsm_arr.shape[0]
        cols = dsm_arr.shape[1]

        dsm_arr[dsm_arr == dsm_nd_val] = 0.0
        if dsm_arr.min() < 0:
            dsmraise = np.abs(dsm_arr.min())
            dsm_arr = dsm_arr + dsmraise
        else:
            dsmraise = 0

        alt = np.median(dsm_arr)
        if alt < 0:
            alt = 3

        # Vegetation
        transVeg = self.params.Tree_settings.Value.Transmissivity
        trunkratio = self.params.Tree_settings.Value.Trunk_ratio
        if self.config.use_veg_dem:
            vegdsm, _, _, _ = common.load_raster(self.config.cdsm_path, bbox=None)
            if self.config.tdsm_path:
                vegdsm2, _, _, _ = common.load_raster(self.config.tdsm_path, bbox=None)
            else:
                vegdsm2 = vegdsm * trunkratio
        else:
            vegdsm = None
            vegdsm2 = None

        # Land cover
        if self.config.use_landcover:
            lcgrid, _, _, _ = common.load_raster(self.config.lc_path, bbox=None)
        else:
            lcgrid = None

        # DEM for buildings
        if self.config.use_dem_for_buildings:
            dem, _, _, dem_nd_val = common.load_raster(self.config.dem_path, bbox=None)
            dem[dem == dem_nd_val] = 0.0
            if dem.min() < 0:
                demraise = np.abs(dem.min())
                dem = dem + demraise
            else:
                demraise = 0

        # SVF
        with zipfile.ZipFile(self.config.svf_path, "r") as zip_ref:
            zip_ref.extractall(self.config.working_dir)
        # Load SVF data
        svf_data = SvfData(self.config.working_dir, self.config.use_veg_dem)
        tmp = svf_data.svf + svf_data.svf_veg - 1.0
        tmp[tmp < 0.0] = 0.0
        svfalfa = np.arcsin(np.exp(np.log(1.0 - tmp) / 2.0))

        wallheight, _, _, _ = common.load_raster(self.config.wh_path, bbox=None)
        wallaspect, _, _, _ = common.load_raster(self.config.wa_path, bbox=None)

        # weather data
        if self.config.use_epw_file:
            weather_data = self.load_epw_weather()
        else:
            weather_data = self.load_met_weather(header_rows=1, delim=" ")

        location = {"longitude": lng, "latitude": lat, "altitude": alt}
        weather_date_arr = weather_data.to_date_arr()
        YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax = Solweig_2015a_metdata_noload(
            weather_date_arr, location, int(self.config.utc)
        )

        # POIs check
        if self.config.poi_path:
            self.load_poi_data(dsm_trf_arr)

        # Posture settings
        if self.params.Tmrt_params.Value.posture == "Standing":
            posture = self.params.Posture.Standing.Value
        else:
            posture = self.params.Posture.Sitting.Value

        Fside, Fup, height, Fcyl = posture.Fside, posture.Fup, posture.height, posture.Fcyl

        # Radiative surface influence
        first = np.round(height)
        if first == 0.0:
            first = 1.0
        second = np.round(height * 20.0)

        if self.config.use_veg_dem:
            # Conifer or deciduous
            if self.config.conifer:
                leafon = np.ones((1, weather_data.DOY.shape[0]))
            else:
                leafon = np.zeros((1, weather_data.DOY.shape[0]))
                if self.params.Tree_settings.Value.First_day_leaf > self.params.Tree_settings.Value.Last_day_leaf:
                    leaf_bool = (self.params.Tree_settings.Value.First_day_leaf < weather_data.DOY) | (
                        self.params.Tree_settings.Value.Last_day_leaf > weather_data.DOY
                    )
                else:
                    leaf_bool = (self.params.Tree_settings.Value.First_day_leaf < weather_data.DOY) & (
                        self.params.Tree_settings.Value.Last_day_leaf > weather_data.DOY
                    )
                leafon[0, leaf_bool] = 1

            # % Vegetation transmittivity of shortwave radiation
            psi = leafon * transVeg
            psi[leafon == 0] = 0.5
            # amaxvalue
            vegmax = vegdsm.max()
            amaxvalue = dsm_arr.max() - dsm_arr.min()
            amaxvalue = np.maximum(amaxvalue, vegmax)

            # Elevation vegdsms if buildingDEM includes ground heights
            vegdsm = vegdsm + dsm_arr
            vegdsm[vegdsm == dsm_arr] = 0
            vegdsm2 = vegdsm2 + dsm_arr
            vegdsm2[vegdsm2 == dsm_arr] = 0

            # % Bush separation
            bush = np.logical_not(vegdsm2 * vegdsm) * vegdsm

            svfbuveg = svf_data.svf - (1.0 - svf_data.svf_veg) * (1.0 - transVeg)  # % major bug fixed 20141203
        else:
            psi = leafon * 0.0 + 1.0
            svfbuveg = svf_data.svf
            bush = np.zeros([rows, cols])
            amaxvalue = 0

        # Initialization of maps
        Knight = np.zeros((rows, cols))
        Tgmap1 = np.zeros((rows, cols))
        Tgmap1E = np.zeros((rows, cols))
        Tgmap1S = np.zeros((rows, cols))
        Tgmap1W = np.zeros((rows, cols))
        Tgmap1N = np.zeros((rows, cols))

        # Create building boolean raster from either land cover or height rasters
        if not self.config.use_dem_for_buildings:
            buildings = np.copy(lcgrid)
            buildings[buildings == 7] = 1
            buildings[buildings == 6] = 1
            buildings[buildings == 5] = 1
            buildings[buildings == 4] = 1
            buildings[buildings == 3] = 1
            buildings[buildings == 2] = 0
        else:
            buildings = dsm_arr - dem
            buildings[buildings < 2.0] = 1.0
            buildings[buildings >= 2.0] = 0.0

        if self.config.save_buildings:
            common.save_raster(
                self.config.output_dir + "/buildings.tif",
                buildings,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
            )

        # Import shadow matrices (Anisotropic sky)
        if self.config.use_aniso:
            data = np.load(self.config.aniso_path)
            shmat = data["shadowmat"]
            vegshmat = data["vegshadowmat"]
            vbshvegshmat = data["vbshmat"]
            if self.config.use_veg_dem:
                diffsh = np.zeros((rows, cols, shmat.shape[2]))
                for i in range(0, shmat.shape[2]):
                    diffsh[:, :, i] = shmat[:, :, i] - (1 - vegshmat[:, :, i]) * (
                        1 - transVeg
                    )  # changes in psi not implemented yet
            else:
                diffsh = shmat

            # Estimate number of patches based on shadow matrices
            if shmat.shape[2] == 145:
                patch_option = 1  # patch_option = 1 # 145 patches
            elif shmat.shape[2] == 153:
                patch_option = 2  # patch_option = 2 # 153 patches
            elif shmat.shape[2] == 306:
                patch_option = 3  # patch_option = 3 # 306 patches
            elif shmat.shape[2] == 612:
                patch_option = 4  # patch_option = 4 # 612 patches

            # asvf to calculate sunlit and shaded patches
            asvf = np.arccos(np.sqrt(svf_data.svf))

            # Empty array for steradians
            steradians = np.zeros(shmat.shape[2])
        else:
            # anisotropic_sky = 0
            diffsh = None
            shmat = None
            vegshmat = None
            vbshvegshmat = None
            asvf = None
            patch_option = 0
            steradians = 0

        # % Ts parameterisation maps
        if self.config.use_landcover:
            # Get land cover properties for Tg wave (land cover scheme based on Bogren et al. 2000, explained in Lindberg et al., 2008 and Lindberg, Onomura & Grimmond, 2016)
            [TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST, TmaxLST_wall] = Tgmaps_v1(
                lcgrid, self.params
            )
        else:
            TgK = Knight + self.params.Ts_deg.Value.Cobble_stone_2014a
            Tstart = Knight - self.params.Tstart.Value.Cobble_stone_2014a
            TmaxLST = self.params.TmaxLST.Value.Cobble_stone_2014a
            alb_grid = Knight + self.params.Albedo.Effective.Value.Cobble_stone_2014a
            emis_grid = Knight + self.params.Emissivity.Value.Cobble_stone_2014a
            TgK_wall = self.params.Ts_deg.Value.Walls
            Tstart_wall = self.params.Tstart.Value.Walls
            TmaxLST_wall = self.params.TmaxLST.Value.Walls

        # Import data for wall temperature parameterization
        if self.config.use_wall_scheme:
            wallData = np.load(self.config.wall_path)
            voxelMaps = wallData["voxelId"]
            voxelTable = wallData["voxelTable"]

            # Get wall type
            if self.qgis_env is False:
                wall_type_standalone = {"Brick_wall": "100", "Concrete_wall": "101", "Wood_wall": "102"}
                wall_type = wall_type_standalone[self.config.wall_type]
            else:
                wall_type = self.config.wall_type

            # Get heights of walls including corners
            walls_scheme = wa.findwalls_sp(dsm_arr, 2, np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]]))
            # Get aspects of walls including corners
            dirwalls_scheme = wa.filter1Goodwin_as_aspect_v3(walls_scheme.copy(), scale, dsm_arr, None, 100.0 / 180.0)
            # Calculate timeStep
            first_timestep = (
                pd.to_datetime(YYYY[0][0], format="%Y")
                + pd.to_timedelta(weather_data.DOY[0] - 1, unit="d")
                + pd.to_timedelta(weather_data.hours[0], unit="h")
                + pd.to_timedelta(weather_data.minu[0], unit="m")
            )
            second_timestep = (
                pd.to_datetime(YYYY[0][1], format="%Y")
                + pd.to_timedelta(weather_data.DOY[1] - 1, unit="d")
                + pd.to_timedelta(weather_data.hours[1], unit="h")
                + pd.to_timedelta(weather_data.minu[1], unit="m")
            )
            timeStep = (second_timestep - first_timestep).seconds

            # Load voxelTable as Pandas DataFrame
            voxelTable, dirwalls_scheme = load_walls(
                voxelTable,
                self.params,
                wall_type,
                dirwalls_scheme,
                weather_data.Ta[0],
                timeStep,
                alb_grid,
                self.config.use_landcover,
                lcgrid,
                dsm_arr,
            )

            # Use wall of interest
            if self.config.woi_path:
                self.load_woi_data(dsm_trf_arr)

            # Create pandas datetime object for NetCDF output
            if self.config.wall_netcdf:
                met_for_xarray = (
                    pd.to_datetime(YYYY[0][:], format="%Y")
                    + pd.to_timedelta(weather_data.DOY - 1, unit="d")
                    + pd.to_timedelta(weather_data.hours, unit="h")
                    + pd.to_timedelta(weather_data.minu, unit="m")
                )
        else:
            voxelMaps = None
            voxelTable = None
            timeStep = 0
            walls_scheme = np.ones((rows, cols)) * 10.0
            dirwalls_scheme = np.ones((rows, cols)) * 10.0

        # Initialisation of time related variables
        if weather_data.Ta.__len__() == 1:
            timestepdec = 0
        else:
            timestepdec = dectime[1] - dectime[0]
        timeadd = 0.0
        firstdaytime = 1.0

        # Save hemispheric image
        if self.config.use_aniso and self.poi_pixel_xys is not None:
            patch_characteristics = hemispheric_image(
                self.poi_pixel_xys,
                shmat,
                vegshmat,
                vbshvegshmat,
                voxelMaps,
                walls_scheme,
            )

        # If metfile starts at night
        CI = 1.0

        # reading variables from config and parameters that is not yet presented
        albedo_b = self.params.Albedo.Effective.Value.Walls
        ewall = self.params.Emissivity.Value.Walls
        elvis = 0.0
        absK = self.params.Tmrt_params.Value.absK
        absL = self.params.Tmrt_params.Value.absL

        # Main loop
        tmrtplot = np.zeros((rows, cols))
        TgOut1 = np.zeros((rows, cols))

        # Initiate array for I0 values
        if np.unique(weather_data.DOY).shape[0] > 1:
            unique_days = np.unique(weather_data.DOY)
            first_unique_day = weather_data.DOY[unique_days[0] == weather_data.DOY]
            I0_array = np.zeros(first_unique_day.shape[0])
        else:
            first_unique_day = weather_data.DOY.copy()
            I0_array = np.zeros(weather_data.DOY.shape[0])

        self.prep_progress(weather_data.Ta.__len__())

        for i in np.arange(0, self.iters_total):
            proceed = self.iter_progress()
            if not proceed:
                break

            # Daily water body temperature
            Twater = []
            if self.config.use_landcover:
                if (dectime[i] - np.floor(dectime[i])) == 0 or (i == 0):
                    Twater = np.mean(weather_data.Ta[jday[0] == np.floor(dectime[i])])

            # Nocturnal cloudfraction from Offerle et al. 2003
            if (dectime[i] - np.floor(dectime[i])) == 0:
                daylines = np.where(np.floor(dectime) == dectime[i])
                if daylines.__len__() > 1:
                    alt_day = altitude[0][daylines]
                    alt2 = np.where(alt_day > 1)
                    rise = alt2[0][0]
                    [_, CI, _, _, _] = clearnessindex_2013b(
                        zen[0, i + rise + 1],
                        jday[0, i + rise + 1],
                        weather_data.Ta[i + rise + 1],
                        weather_data.RH[i + rise + 1] / 100.0,
                        weather_data.radG[i + rise + 1],
                        location,
                        weather_data.P[i + rise + 1],
                    )
                    if (CI > 1.0) or (np.inf == CI):
                        CI = 1.0
                else:
                    CI = 1.0

            (
                Tmrt,
                Kdown,
                Kup,
                Ldown,
                Lup,
                Tg,
                ea,
                esky,
                I0,
                CI,
                shadow,
                firstdaytime,
                timestepdec,
                timeadd,
                Tgmap1,
                Tgmap1E,
                Tgmap1S,
                Tgmap1W,
                Tgmap1N,
                Keast,
                Ksouth,
                Kwest,
                Knorth,
                Least,
                Lsouth,
                Lwest,
                Lnorth,
                KsideI,
                TgOut1,
                TgOut,
                radIout,
                radDout,
                Lside,
                Lsky_patch_characteristics,
                CI_Tg,
                CI_TgG,
                KsideD,
                dRad,
                Kside,
                steradians,
                voxelTable,
            ) = so.Solweig_2025a_calc(
                i,
                dsm_arr,
                scale,
                rows,
                cols,
                svf_data.svf,
                svf_data.svf_north,
                svf_data.svf_west,
                svf_data.svf_east,
                svf_data.svf_south,
                svf_data.svf_veg,
                svf_data.svf_veg_north,
                svf_data.svf_veg_east,
                svf_data.svf_veg_south,
                svf_data.svf_veg_west,
                svf_data.svf_veg_blocks_bldg_sh,
                svf_data.svf_veg_blocks_bldg_sh_east,
                svf_data.svf_veg_blocks_bldg_sh_south,
                svf_data.svf_veg_blocks_bldg_sh_west,
                svf_data.svf_veg_blocks_bldg_sh_north,
                vegdsm,
                vegdsm2,
                albedo_b,
                absK,
                absL,
                ewall,
                Fside,
                Fup,
                Fcyl,
                altitude[0][i],
                azimuth[0][i],
                zen[0][i],
                jday[0][i],
                self.config.use_veg_dem,
                self.config.only_global,
                buildings,
                location,
                psi[0][i],
                self.config.use_landcover,
                lcgrid,
                dectime[i],
                altmax[0][i],
                wallaspect,
                wallheight,
                self.config.person_cylinder,
                elvis,
                weather_data.Ta[i],
                weather_data.RH[i],
                weather_data.radG[i],
                weather_data.radD[i],
                weather_data.radI[i],
                weather_data.P[i],
                amaxvalue,
                bush,
                Twater,
                TgK,
                Tstart,
                alb_grid,
                emis_grid,
                TgK_wall,
                Tstart_wall,
                TmaxLST,
                TmaxLST_wall,
                first,
                second,
                svfalfa,
                svfbuveg,
                firstdaytime,
                timeadd,
                timestepdec,
                Tgmap1,
                Tgmap1E,
                Tgmap1S,
                Tgmap1W,
                Tgmap1N,
                CI,
                TgOut1,
                diffsh,
                shmat,
                vegshmat,
                vbshvegshmat,
                self.config.use_aniso,
                asvf,
                patch_option,
                voxelMaps,
                voxelTable,
                weather_data.Ws[i],
                self.config.use_wall_scheme,
                timeStep,
                steradians,
                walls_scheme,
                dirwalls_scheme,
            )

            # Save I0 for I0 vs. Kdown output plot to check if UTC is off
            if i < first_unique_day.shape[0]:
                I0_array[i] = I0
            elif i == first_unique_day.shape[0]:
                # Output I0 vs. Kglobal plot
                radG_for_plot = weather_data.radG[first_unique_day[0] == weather_data.DOY]
                dectime_for_plot = dectime[first_unique_day[0] == weather_data.DOY]
                fig, ax = plt.subplots()
                ax.plot(dectime_for_plot, I0_array, label="I0")
                ax.plot(dectime_for_plot, radG_for_plot, label="Kglobal")
                ax.set_ylabel("Shortwave radiation [$Wm^{-2}$]")
                ax.set_xlabel("Decimal time")
                ax.set_title("UTC" + str(self.config.utc))
                ax.legend()
                fig.savefig(self.config.output_dir + "/metCheck.png", dpi=150)

            tmrtplot = tmrtplot + Tmrt

            if altitude[0][i] > 0:
                w = "D"
            else:
                w = "N"

            if weather_data.hours[i] < 10:
                XH = "0"
            else:
                XH = ""

            if weather_data.minu[i] < 10:
                XM = "0"
            else:
                XM = ""

            if self.poi_pixel_xys is not None:
                for n in range(0, self.poi_pixel_xys.shape[0]):
                    idx, row_idx, col_idx = self.poi_pixel_xys[n]
                    row_idx = int(row_idx)
                    col_idx = int(col_idx)
                    result_row = {
                        "poi_idx": idx,
                        "yyyy": YYYY[0, i],
                        "id": jday[0, i],
                        "it": weather_data.hours[i],
                        "imin": weather_data.minu[i],
                        "dectime": dectime[i],
                        "altitude": altitude[0, i],
                        "azimuth": azimuth[0, i],
                        "kdir": radIout,
                        "kdiff": radDout,
                        "kglobal": weather_data.radG[i],
                        "kdown": Kdown[row_idx, col_idx],
                        "kup": Kup[row_idx, col_idx],
                        "keast": Keast[row_idx, col_idx],
                        "ksouth": Ksouth[row_idx, col_idx],
                        "kwest": Kwest[row_idx, col_idx],
                        "knorth": Knorth[row_idx, col_idx],
                        "ldown": Ldown[row_idx, col_idx],
                        "lup": Lup[row_idx, col_idx],
                        "least": Least[row_idx, col_idx],
                        "lsouth": Lsouth[row_idx, col_idx],
                        "lwest": Lwest[row_idx, col_idx],
                        "lnorth": Lnorth[row_idx, col_idx],
                        "Ta": weather_data.Ta[i],
                        "Tg": TgOut[row_idx, col_idx],
                        "RH": weather_data.RH[i],
                        "Esky": esky,
                        "Tmrt": Tmrt[row_idx, col_idx],
                        "I0": I0,
                        "CI": CI,
                        "Shadow": shadow[row_idx, col_idx],
                        "SVF_b": svf_data.svf[row_idx, col_idx],
                        "SVF_bv": svfbuveg[row_idx, col_idx],
                        "KsideI": KsideI[row_idx, col_idx],
                    }
                    # Recalculating wind speed based on powerlaw
                    WsPET = (1.1 / self.params.Wind_Height.Value.magl) ** 0.2 * weather_data.Ws[i]
                    WsUTCI = (10.0 / self.params.Wind_Height.Value.magl) ** 0.2 * weather_data.Ws[i]
                    resultPET = PET_calculations._PET(
                        weather_data.Ta[i],
                        weather_data.RH[i],
                        Tmrt[row_idx, col_idx],
                        WsPET,
                        self.params.PET_settings.Value.Weight,
                        self.params.PET_settings.Value.Age,
                        self.params.PET_settings.Value.Height,
                        self.params.PET_settings.Value.Activity,
                        self.params.PET_settings.Value.clo,
                        self.params.PET_settings.Value.Sex,
                    )
                    result_row["PET"] = resultPET
                    resultUTCI = utci.utci_calculator(
                        weather_data.Ta[i], weather_data.RH[i], Tmrt[row_idx, col_idx], WsUTCI
                    )
                    result_row["UTCI"] = resultUTCI
                    result_row["CI_Tg"] = CI_Tg
                    result_row["CI_TgG"] = CI_TgG
                    result_row["KsideD"] = KsideD[row_idx, col_idx]
                    result_row["Lside"] = Lside[row_idx, col_idx]
                    result_row["diffDown"] = dRad[row_idx, col_idx]
                    result_row["Kside"] = Kside[row_idx, col_idx]
                    self.poi_results.append(result_row)

            if self.woi_pixel_xys is not None:
                for n in range(0, self.woi_pixel_xys.shape[0]):
                    idx, row_idx, col_idx = self.woi_pixel_xys[n]
                    row_idx = int(row_idx)
                    col_idx = int(col_idx)

                    temp_wall = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "wallTemperature"
                    ].to_numpy()
                    K_in = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "K_in"
                    ].to_numpy()
                    L_in = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "L_in"
                    ].to_numpy()
                    wallShade = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "wallShade"
                    ].to_numpy()

                    result_row = {
                        "woi_idx": idx,
                        "woi_name": self.woi_names[idx],
                        "yyyy": YYYY[0, i],
                        "id": jday[0, i],
                        "it": weather_data.hours[i],
                        "imin": weather_data.minu[i],
                        "dectime": dectime[i],
                        "Ta": weather_data.Ta[i],
                        "SVF": svf_data.svf[row_idx, col_idx],
                        "Ts": temp_wall,
                        "Kin": K_in,
                        "Lin": L_in,
                        "shade": wallShade,
                        "pixel_x": col_idx,
                        "pixel_y": row_idx,
                    }
                    self.woi_results.append(result_row)

                if self.config.wall_netcdf:
                    netcdf_output = self.config.output_dir + "/walls.nc"
                    walls_as_netcdf(
                        voxelTable, rows, cols, met_for_xarray, i, dsm_arr, self.config.dsm_path, netcdf_output
                    )

            time_code = (
                str(int(YYYY[0, i]))
                + "_"
                + str(int(weather_data.DOY[i]))
                + "_"
                + XH
                + str(int(weather_data.hours[i]))
                + XM
                + str(int(weather_data.minu[i]))
                + w
            )

            if self.config.output_tmrt:
                common.save_raster(
                    self.config.output_dir + "/Tmrt_" + time_code + ".tif",
                    Tmrt,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_kup:
                common.save_raster(
                    self.config.output_dir + "/Kup_" + time_code + ".tif",
                    Kup,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_kdown:
                common.save_raster(
                    self.config.output_dir + "/Kdown_" + time_code + ".tif",
                    Kdown,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_lup:
                common.save_raster(
                    self.config.output_dir + "/Lup_" + time_code + ".tif",
                    Lup,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_ldown:
                common.save_raster(
                    self.config.output_dir + "/Ldown_" + time_code + ".tif",
                    Ldown,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_sh:
                common.save_raster(
                    self.config.output_dir + "/Shadow_" + time_code + ".tif",
                    shadow,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_kdiff:
                common.save_raster(
                    self.config.output_dir + "/Kdiff_" + time_code + ".tif",
                    dRad,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )

            # Sky view image of patches
            if (self.config.aniso_path) and (i == 0) and (self.poi_pixel_xys is not None):
                for k in range(self.poi_pixel_xys.shape[0]):
                    Lsky_patch_characteristics[:, 2] = patch_characteristics[:, k]
                    skyviewimage_out = self.config.output_dir + "/POI_" + str(self.poi_names[k]) + ".png"
                    PolarBarPlot(
                        Lsky_patch_characteristics,
                        altitude[0][i],
                        azimuth[0][i],
                        "Hemisphere partitioning",
                        skyviewimage_out,
                        0,
                        5,
                        0,
                    )

        # Save POI results
        if self.poi_results:
            self.save_poi_results(dsm_trf_arr, dsm_crs_wkt)

        # Save WOI results
        if self.woi_results:
            self.save_woi_results(dsm_trf_arr, dsm_crs_wkt)

        # Save Tree Planter results
        if self.config.output_tree_planter:
            albedo_g = self.params.Albedo.Effective.Value.Cobble_stone_2014a
            eground = self.params.Emissivity.Value.Cobble_stone_2014a
            pos = 1 if self.params.Tmrt_params.Value.posture == "Standing" else 0

            settingsHeader = "UTC, posture, onlyglobal, landcover, anisotropic, cylinder, albedo_walls, albedo_ground, emissivity_walls, emissivity_ground, absK, absL, elevation, patch_option"
            settingsFmt = (
                "%i",
                "%i",
                "%i",
                "%i",
                "%i",
                "%i",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%i",
            )
            settingsData = np.array(
                [
                    [
                        int(self.config.utc),
                        pos,
                        self.config.only_global,
                        self.config.use_landcover,
                        self.config.use_aniso,
                        self.config.person_cylinder,
                        albedo_b,
                        albedo_g,
                        ewall,
                        eground,
                        absK,
                        absL,
                        alt,
                        patch_option,
                    ]
                ]
            )
            np.savetxt(
                self.config.output_dir + "/treeplantersettings.txt",
                settingsData,
                fmt=settingsFmt,
                header=settingsHeader,
                delimiter=" ",
            )

        # Save average Tmrt raster
        tmrtplot = tmrtplot / self.iters_total
        common.save_raster(
            self.config.output_dir + "/Tmrt_average.tif",
            tmrtplot,
            dsm_trf_arr,
            dsm_crs_wkt,
            dsm_nd_val,
        )



import numpy as np
from copy import deepcopy
from . import emissivity_models
from . import patch_characteristics

''' This function combines the method to divide the sky vault into patches (Tregenza (1987) and Robinson & Stone (2004)) 
    and the approach by Unsworth & Monteith or Martin & Berdahl (1984) or Bliss (1961) to calculate emissivities of the 
    different parts of the sky vault. '''

def Lcyl_v2022a(esky, sky_patches, Ta, Tgwall, ewall, Lup, shmat, vegshmat, vbshvegshmat, solar_altitude, solar_azimuth, rows, cols, asvf, current_step):

    # Stefan-Boltzmann's Constant
    SBC = 5.67051e-8

    # Sky longwave radiation from emissivity based on Prata (1996)
    Ldown_prata = (esky * SBC * ((Ta + 273.15) ** 4))

    # Degrees to radians
    deg2rad = np.pi / 180

    # Unique altitudes for patches
    skyalt, skyalt_c = np.unique(sky_patches[:, 0], return_counts=True)
    # skyzen = 90-skyalt                  # Unique zeniths for the patches

    # Altitudes of the Robinson & Stone patches
    patch_altitude = sky_patches[:, 0]
    # Azimuths of the Robinson & Stone patches, used for box
    patch_azimuth = sky_patches[:, 1]

    emis_m = 2

    # Unsworth & Monteith (1975)
    if emis_m == 1:
        patch_emissivity_normalized, esky_band = emissivity_models.model1(sky_patches, esky, Ta)
    # Martin & Berdahl (1984)
    elif emis_m == 2:
        patch_emissivity_normalized, esky_band = emissivity_models.model2(sky_patches, esky, Ta)
    # Bliss (1961)
    elif emis_m == 3:
        patch_emissivity_normalized, esky_band = emissivity_models.model3(sky_patches, esky, Ta)

    # Calculation of steradian for each patch
    steradian = np.zeros((patch_altitude.shape[0]))
    for i in range(patch_altitude.shape[0]):
        # If there are more than one patch in a band
        if skyalt_c[skyalt == patch_altitude[i]] > 1:
            steradian[i] = ((360 / skyalt_c[skyalt == patch_altitude[i]]) * deg2rad) * (np.sin((patch_altitude[i] + patch_altitude[0]) * deg2rad) \
            - np.sin((patch_altitude[i] - patch_altitude[0]) * deg2rad))
        # If there is only one patch in band, i.e. 90 degrees
        else:
            steradian[i] = ((360 / skyalt_c[skyalt == patch_altitude[i]]) * deg2rad) * (np.sin((patch_altitude[i]) * deg2rad) \
                - np.sin((patch_altitude[i-1] + patch_altitude[0]) * deg2rad))

    # True = anisotropic sky, False = isotropic sky
    anisotropic_sky = True
    # anisotropic_sky = False

    # Longwave based on spectral flux density (divide by pi)
    Ldown = np.zeros((patch_altitude.shape[0]))
    Lside = np.zeros((patch_altitude.shape[0]))
    Lnormal = np.zeros((patch_altitude.shape[0]))
    for altitude in skyalt:
        # Anisotropic sky
        if anisotropic_sky:
            temp_emissivity = esky_band[skyalt == altitude]
        # Isotropic sky but with patches (need to switch anisotropic_sky to False)
        else:
            temp_emissivity = esky
        # Estimate longwave radiation on a horizontal surface (Ldown), vertical surface (Lside) and perpendicular (Lnormal)
        Ldown[patch_altitude == altitude] = ((temp_emissivity * SBC * ((Ta + 273.15) ** 4)) / np.pi) * steradian[patch_altitude == altitude] * np.sin(altitude * deg2rad)
        Lside[patch_altitude == altitude] = ((temp_emissivity * SBC * ((Ta + 273.15) ** 4)) / np.pi) * steradian[patch_altitude == altitude] * np.cos(altitude * deg2rad)
        Lnormal[patch_altitude == altitude] = ((temp_emissivity * SBC * ((Ta + 273.15) ** 4)) / np.pi) * steradian[patch_altitude == altitude]

    Lsky_normal = deepcopy(sky_patches)
    Lsky_down = deepcopy(sky_patches)
    Lsky_side = deepcopy(sky_patches)

    Lsky_normal[:,2] = Lnormal
    Lsky_down[:,2] = Ldown
    Lsky_side[:,2] = Lside

    # Estimate longwave radiation in each patch based on patch characteristics, i.e. sky, vegetation or building (shaded or sunlit)
    Ldown, Lside, Lside_sky, Lside_veg, Lside_sh, Lside_sun, Lside_ref, \
            Least_, Lwest_, Lnorth_, Lsouth_ = patch_characteristics.define_patch_characteristics(solar_altitude, solar_azimuth, 
                                 patch_altitude, patch_azimuth, steradian,
                                 asvf,
                                 shmat, vegshmat, vbshvegshmat,
                                 Lsky_down, Lsky_side, Lsky_normal, Lup,
                                 Ta, Tgwall, ewall,
                                 rows, cols, current_step)

    # print('Lside_sky old = ' + str(Lside_sky.max()))
    # print('Lside_veg old = ' + str(Lside_veg.max()))
    # print('Lside_sh old = ' + str(Lside_sh.max()))
    # print('Lside_sun old = ' + str(Lside_sun.max()))
    # print('Lside_ref old = ' + str(Lside_ref.max()))

    return Ldown, Lside, Least_, Lwest_, Lnorth_, Lsouth_
    # return Ldown, Lside, Lside_sky, Lside_veg, Lside_sh, Lside_sun, Lside_ref, Lsky_normal, Lsky_down, Lsky_side, Least_, Lwest_, Lnorth_, Lsouth_


import numpy as np
from . import sunlit_shaded_patches

def shortwave_from_sky(sky, angle_of_incidence, lumChi, steradian, patch_azimuth, cyl):
    '''Calculates the amount of diffuse shortwave radiation from the sky for a patch with:
    angle of incidence = angle_of_incidence
    luminance = lumChi
    steradian = steradian'''

    # Diffuse vertical radiation
    diffuse_shortwave_radiation = sky * lumChi * angle_of_incidence * steradian

    return diffuse_shortwave_radiation

def longwave_from_sky(sky, Lsky_side, Lsky_down, patch_azimuth):

    # Degrees to radians
    deg2rad = np.pi / 180    

    # Longwave radiation from sky to vertical surface
    Ldown_sky = sky * Lsky_down

    # Longwave radiation from sky to horizontal surface
    Lside_sky = sky * Lsky_side

    #
    Least = np.zeros((sky.shape[0], sky.shape[1])) 
    Lsouth = np.zeros((sky.shape[0], sky.shape[1]))
    Lwest = np.zeros((sky.shape[0], sky.shape[1]))
    Lnorth = np.zeros((sky.shape[0], sky.shape[1]))

    # Portion into cardinal directions to be used for standing box or POI output
    if (patch_azimuth > 360) or (patch_azimuth < 180):
        Least = sky * Lsky_side * np.cos((90 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 90) and (patch_azimuth < 270):
        Lsouth = sky * Lsky_side * np.cos((180 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 180) and (patch_azimuth < 360):
        Lwest = sky * Lsky_side * np.cos((270 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 270) or (patch_azimuth < 90):
        Lnorth = sky * Lsky_side * np.cos((0 - patch_azimuth) * deg2rad)

    return Lside_sky, Ldown_sky, Least, Lsouth, Lwest, Lnorth

def longwave_from_veg(vegetation, steradian, angle_of_incidence, angle_of_incidence_h, patch_altitude, patch_azimuth, ewall, Ta):
    '''Calculates the amount of longwave radiation from vegetation for a patch with:
    angle of incidence = angle_of_incidence
    steradian = steradian
    if a patch is vegetation = vegetation
    amount of radiation from vegetated patch = vegetation_surface'''

    # Stefan-Boltzmann's Constant
    SBC = 5.67051e-8

    # Degrees to radians
    deg2rad = np.pi / 180    
    
    # Longwave radiation from vegetation surface (considered vertical)
    vegetation_surface = ((ewall * SBC * ((Ta + 273.15) ** 4)) / np.pi)

    # Longwave radiation reaching a vertical surface
    Lside_veg = vegetation_surface * steradian * angle_of_incidence * vegetation

    # Longwave radiation reaching a horizontal surface
    Ldown_veg = vegetation_surface * steradian * angle_of_incidence_h * vegetation

    #
    Least = np.zeros((vegetation.shape[0], vegetation.shape[1])) 
    Lsouth = np.zeros((vegetation.shape[0], vegetation.shape[1]))
    Lwest = np.zeros((vegetation.shape[0], vegetation.shape[1]))
    Lnorth = np.zeros((vegetation.shape[0], vegetation.shape[1]))

    # Portion into cardinal directions to be used for standing box or POI output
    if (patch_azimuth > 360) or (patch_azimuth < 180):
        Least = vegetation_surface * steradian * np.cos(patch_altitude * deg2rad) * vegetation * np.cos((90 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 90) and (patch_azimuth < 270):
        Lsouth = vegetation_surface * steradian * np.cos(patch_altitude * deg2rad) * vegetation * np.cos((180 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 180) and (patch_azimuth < 360):
        Lwest = vegetation_surface * steradian * np.cos(patch_altitude * deg2rad) * vegetation * np.cos((270 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 270) or (patch_azimuth < 90):
        Lnorth = vegetation_surface * steradian * np.cos(patch_altitude * deg2rad) * vegetation * np.cos((0 - patch_azimuth) * deg2rad)    

    return Lside_veg, Ldown_veg, Least, Lsouth, Lwest, Lnorth

def longwave_from_buildings(building, steradian, angle_of_incidence, angle_of_incidence_h, patch_azimuth, sunlit_patches, shaded_patches, azimuth_difference, solar_altitude, ewall, Ta, Tgwall):

    # Stefan-Boltzmann's Constant
    SBC = 5.67051e-8

    # Degrees to radians
    deg2rad = np.pi / 180    

    #
    Least = np.zeros((building.shape[0], building.shape[1])) 
    Lsouth = np.zeros((building.shape[0], building.shape[1]))
    Lwest = np.zeros((building.shape[0], building.shape[1]))
    Lnorth = np.zeros((building.shape[0], building.shape[1]))

    # Longwave radiation from sunlit surfaces
    sunlit_surface = ((ewall * SBC * ((Ta + Tgwall + 273.15) ** 4)) / np.pi)
    # Longwave radiation from shaded surfaces
    shaded_surface = ((ewall * SBC * ((Ta + 273.15) ** 4)) / np.pi)
    if ((azimuth_difference > 90) and (azimuth_difference < 270) and (solar_altitude > 0)):
        # Calculate which patches defined as buildings that are sunlit or shaded
        # sunlit_patches, shaded_patches = sunlit_shaded_patches.shaded_or_sunlit(solar_altitude, solar_azimuth, patch_altitude, patch_azimuth, asvf)
        
        # Calculate longwave radiation from sunlit walls to vertical surface
        Lside_sun = sunlit_surface * sunlit_patches * steradian * angle_of_incidence * building
        # Calculate longwave radiation from shaded walls to vertical surface
        Lside_sh = shaded_surface * shaded_patches * steradian * angle_of_incidence * building
        
        # Calculate longwave radiation from sunlit walls to horizontal surface
        Ldown_sun = sunlit_surface * sunlit_patches * steradian * angle_of_incidence_h * building
        # Calculate longwave radiation from shaded walls to horizontal surface
        Ldown_sh = shaded_surface * shaded_patches * steradian * angle_of_incidence_h * building
        
        # Portion into cardinal directions to be used for standing box or POI output
        if (patch_azimuth > 360) or (patch_azimuth < 180):
            Least = sunlit_surface * sunlit_patches * steradian * angle_of_incidence * building * np.cos((90 - patch_azimuth) * deg2rad)
            Least += shaded_surface * shaded_patches * steradian * angle_of_incidence * building * np.cos((90 - patch_azimuth) * deg2rad)
        if (patch_azimuth > 90) and (patch_azimuth < 270):
            Lsouth = sunlit_surface * sunlit_patches * steradian * angle_of_incidence * building * np.cos((180 - patch_azimuth) * deg2rad)
            Lsouth += shaded_surface * shaded_patches * steradian * angle_of_incidence * building * np.cos((180 - patch_azimuth) * deg2rad)
        if (patch_azimuth > 180) and (patch_azimuth < 360):
            Lwest = sunlit_surface * sunlit_patches * steradian * angle_of_incidence * building * np.cos((270 - patch_azimuth) * deg2rad)
            Lwest += shaded_surface * shaded_patches * steradian * angle_of_incidence * building * np.cos((270 - patch_azimuth) * deg2rad)
        if (patch_azimuth > 270) or (patch_azimuth < 90):
            Lnorth = sunlit_surface * sunlit_patches * steradian * angle_of_incidence * building * np.cos((0 - patch_azimuth) * deg2rad)
            Lnorth += shaded_surface * shaded_patches * steradian * angle_of_incidence * building * np.cos((0 - patch_azimuth) * deg2rad)

    else:
        # Calculate longwave radiation from shaded walls reaching a vertical surface
        Lside_sh = shaded_surface * steradian * angle_of_incidence * building
        Lside_sun = np.zeros((Lside_sh.shape[0], Lside_sh.shape[1]))
        
        # Calculate longwave radiation from shaded walls reaching a horizontal surface
        Ldown_sh = shaded_surface * steradian * angle_of_incidence_h * building
        Ldown_sun = np.zeros((Lside_sh.shape[0], Lside_sh.shape[1]))

        # Portion into cardinal directions to be used for standing box or POI output
        if (patch_azimuth > 360) or (patch_azimuth < 180):
            Least = shaded_surface * steradian * angle_of_incidence * building * np.cos((90 - patch_azimuth) * deg2rad)
        if (patch_azimuth > 90) and (patch_azimuth < 270):
            Lsouth = shaded_surface * steradian * angle_of_incidence * building * np.cos((180 - patch_azimuth) * deg2rad)
        if (patch_azimuth > 180) and (patch_azimuth < 360):
            Lwest = shaded_surface * steradian * angle_of_incidence * building * np.cos((270 - patch_azimuth) * deg2rad)
        if (patch_azimuth > 270) or (patch_azimuth < 90):
            Lnorth = shaded_surface * steradian * angle_of_incidence * building * np.cos((0 - patch_azimuth) * deg2rad)

    return Lside_sun, Lside_sh, Ldown_sun, Ldown_sh, Least, Lsouth, Lwest, Lnorth

def longwave_from_buildings_wallScheme(voxelMaps, voxelTable, steradian, angle_of_incidence, angle_of_incidence_h, patch_azimuth):

    # Degrees to radians
    deg2rad = np.pi / 180    

    #
    Lside = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))
    Lside_sh = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))
    Ldown = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))
    Ldown_sh = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))
    Least = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1])) 
    Lsouth = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))
    Lwest = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))
    Lnorth = np.zeros((voxelMaps.shape[0], voxelMaps.shape[1]))    
    
    # print(voxelMaps)
    #print(voxelTable.head())
    l = list(np.unique(voxelMaps)[1:])
    # print(l)
    a = dict(voxelTable.loc[l, 'LongwaveRadiation'])
    # print(a)
    patch_radiation = np.vectorize(a.get)(voxelMaps).astype(float)
    patch_radiation[np.isnan(patch_radiation)] = 0
    Lside += patch_radiation * steradian * angle_of_incidence
    Ldown += patch_radiation * steradian * angle_of_incidence_h

    # Portion into cardinal directions to be used for standing box or POI output
    if (patch_azimuth > 360) or (patch_azimuth < 180):
        Least = patch_radiation * steradian * angle_of_incidence * np.cos((90 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 90) and (patch_azimuth < 270):
        Lsouth = patch_radiation * steradian * angle_of_incidence * np.cos((180 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 180) and (patch_azimuth < 360):
        Lwest = patch_radiation * steradian * angle_of_incidence * np.cos((270 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 270) or (patch_azimuth < 90):
        Lnorth = patch_radiation * steradian * angle_of_incidence * np.cos((0 - patch_azimuth) * deg2rad)    

    return Lside, Lside_sh, Ldown, Ldown_sh, Least, Lsouth, Lwest, Lnorth

def reflected_longwave(reflecting_surface, steradian, angle_of_incidence, angle_of_incidence_h, patch_azimuth, Ldown_sky, Lup, ewall):

    # Degrees to radians
    deg2rad = np.pi / 180  

    # Calculate reflected longwave in each patch
    reflected_radiation = (((Ldown_sky+Lup) * (1-ewall)*0.5) / np.pi)

    # Reflected longwave radiation reaching vertical surfaces
    Lside_ref = reflected_radiation * steradian * angle_of_incidence * reflecting_surface
    
    # Reflected longwave radiation reaching horizontal surfaces
    Ldown_ref = reflected_radiation * steradian * angle_of_incidence_h * reflecting_surface

    #
    Least = np.zeros((reflecting_surface.shape[0], reflecting_surface.shape[1])) 
    Lsouth = np.zeros((reflecting_surface.shape[0], reflecting_surface.shape[1]))
    Lwest = np.zeros((reflecting_surface.shape[0], reflecting_surface.shape[1]))
    Lnorth = np.zeros((reflecting_surface.shape[0], reflecting_surface.shape[1]))

    # Portion into cardinal directions to be used for standing box or POI output
    if (patch_azimuth > 360) or (patch_azimuth < 180):
        Least = reflected_radiation * steradian * angle_of_incidence * reflecting_surface * np.cos((90 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 90) and (patch_azimuth < 270):
        Lsouth = reflected_radiation * steradian * angle_of_incidence * reflecting_surface * np.cos((180 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 180) and (patch_azimuth < 360):
        Lwest = reflected_radiation * steradian * angle_of_incidence * reflecting_surface * np.cos((270 - patch_azimuth) * deg2rad)
    if (patch_azimuth > 270) or (patch_azimuth < 90):
        Lnorth = reflected_radiation * steradian * angle_of_incidence * reflecting_surface * np.cos((0 - patch_azimuth) * deg2rad)

    return Lside_ref, Ldown_ref, Least, Lsouth, Lwest, Lnorth

def patch_steradians(L_patches):
    ''''This function calculates the steradians of the patches'''

    # Degrees to radians
    deg2rad = np.pi / 180  

    # Unique altitudes for patches
    skyalt, skyalt_c = np.unique(L_patches[:, 0], return_counts=True)

    # Altitudes of the Robinson & Stone patches
    patch_altitude = L_patches[:, 0]

    # Calculation of steradian for each patch
    steradian = np.zeros((patch_altitude.shape[0]))
    for i in range(patch_altitude.shape[0]):
        # If there are more than one patch in a band
        if skyalt_c[skyalt == patch_altitude[i]] > 1:
            steradian[i] = ((360 / skyalt_c[skyalt == patch_altitude[i]]) * deg2rad) * (np.sin((patch_altitude[i] + patch_altitude[0]) * deg2rad) \
            - np.sin((patch_altitude[i] - patch_altitude[0]) * deg2rad))
        # If there is only one patch in band, i.e. 90 degrees
        else:
            steradian[i] = ((360 / skyalt_c[skyalt == patch_altitude[i]]) * deg2rad) * (np.sin((patch_altitude[i]) * deg2rad) \
                - np.sin((patch_altitude[i-1] + patch_altitude[0]) * deg2rad))     
            
    return steradian, skyalt, patch_altitude


import numpy as np

def COMFA_Mact(weight, height, sex, age, activity, activity_unit):
    # Weight of  kid (kg)
    # weight = 20
    # Height of  kid (cm)
    # height = 114
    # Sex/gender of kid, 1 = boy, 2 = girl
    # sex = 1
    # Age of kid
    # age = 6
    # activity (MET), activity if kid
    # activity = 3.8
    # activity time, time of activity in minutes
    # activity_time = 30

    ### METABOLIC HEAT CALCULATION FOR CHILDREN ###
    # Body surface area of child or adult (different for infant according to Haylock et al. (1978))
    BSA = weight**0.5378 * height**0.3964 * 0.0242
    
    if activity_unit == 'MET':

        if sex == 2:
            if ((age >= 3) and (age <= 10)):
                # Resting metabolic rate of a 3-10 year old girl (kJ/day) (Cheng & Brown, 2020)
                RMR_J = 16.97*weight + 1.618*height + 371.2
            elif ((age > 10) and (age <= 18)):
                # Resting metabolic rate of a 10-18 year old girl (kJ/day) (Cheng & Brown, 2020)
                RMR_J = 8.365*weight + 4.65*height + 200
            else:
                RMR_J = 8.365*weight + 4.65*height + 200
        elif sex == 1:
            if ((age >= 3) and (age <= 10)):
                # Resting metabolic rate of a 3-10 year old boy (kJ/day) (Cheng & Brown, 2020)
                RMR_J = 19.6*weight + 1.033*height + 414.9
            elif ((age > 10) and (age <= 18)):
                # Resting metabolic rate of a 10-18 year old boy (kJ/day) (Cheng & Brown, 2020)
                RMR_J = 16.25*weight + 1.372*height + 515.5
            else:
                RMR_J = 8.365*weight + 4.65*height + 200
        # Convert to per hour. Original (kJ/day) = 24 hours.
        RMR_J = RMR_J / 24
        # Resting metabolic rate in Wm-2 for person
        RMR = RMR_J/BSA
        # Total metabolic energy by a person
        Mact = RMR * activity # * activity_time
        # Total metabolic energy by a person for PET
        Mact_PET = Mact * BSA
    elif activity_unit == 'W':
        Mact = activity / BSA
        Mact_PET = activity

    return Mact, Mact_PET

def COMFA_e(Ta, RH):
    # Tetens equation to calculate saturated vapour pressure (kPa) ata given air temperaure, 
    # with inputs Ta (air temperature, degrees C).
    # From Natasha Kenny and re-written by Jenni Vanos, April 2009, translated from MATLAB to Python by Nils Wallenberg 2022.
    # Saturated Vapour Pressure (kPa)
    es = 0.61354 * np.exp((17.502 * Ta) / (240.97 + Ta))
    # Ambient Vapour Pressure (kPa)
    ea = (RH / 100) * es

    # Saturation vapour pressure
    # es = 6.112 * np.exp( ( 17.67 * Ta ) / ( Ta + 243.5 ) )
    # Ambient vapour pressure
    # ea = (RH / 100) * es

    return ea

def COMFA_MET(Mact, Ta, RH):    
    # Ambient vapour pressure
    ea = COMFA_e(Ta, RH)    
    # Heat consumed when breathing
    f = 0.150 - 0.0173 * ea - 0.0014 * Ta

    MET = (1 - f) * Mact

    return MET

def COMFA_rc(va, rco):
    # Calculates the clothing resistance for 
    # input for the COMFA energy balance equation with inputs rco (insulation
    # value of clothing, s/m), va (activity wind velocity in m/s), vw
    # (windspeed, m/s). Note that rco must be known based on clothing worn. 

    if va == 0:
        rc = rco
    else:
        rc = rco * (-0.37 * (1 - np.exp(-va / 0.72)) + 1)
        # rc = rco * (-0.37 * (1 - np.exp(-va / 0.72)) + 1)

    # This is new equation from Kenny et al. (2009b) - IJB.

    # Alternative method for standing individual when clothing isn't known: -
    # UTCI regression equation (Havenith 2012 IJB). 
    # 
    # Calculates the clothing resistance for input for the COMFA energy balance 
    # equation with inputs of Tair (degrees C) to calculate the Icl (clo)based 
    # on air temperature from the UTCI equation. See Vanos, 2012b. 
    # For use on standing individuals only. Can be slowly walking or standing.

    # Written Oct 2009 by Jenni Vanos
    # Because we are assuming little to no activity speed (va) then rc = rco.

    # Icl = 1.372 - (0.01866*Tair) - (0.0004849*(Tair.^2)) - (0.000009333*(Tair.^3)); %Icl here is in clo.
    # Icl equation is from UTCI, and more info can be found in Vanos et al.
    # (2012b). 
    # rc = Icl*186.6; %186.6 is the conversion to put a clo into a s/m.

    return rc

def COMFA_vr(vw, va):
    # Written Oct 2009 by Jenni Vanos
    # This equation calculates the relative windspeed (vr) in m/s using the
    # activity windspeed (va) and windspeed (vw).

    vr = np.sqrt(va**2 + vw**2)

    # To acount for wind direction and activity direction, see Vanos et al.
    # (2012b) using angular equation from ISO for Vr. 

    return vr

def COMFA_ra(vw, va):
    # COMFA equation input under Convective Heat Loss or Gain (CONV)

    # Calculates the resistance of the boundary air layer around the body (s/m) with
    # inputs vr (Relative Wind velocity, m/s)
    # vw (wind speed, m/s), va (activity speed, m/s)

    # Thermal diffusivity of air (m2/s)
    k = 22e-6
    # kinematic viscosity of air (m2/s)
    v = 1.5e-5
    # diameter of cylinder as a model of a human (m)(Campbell, 1977)
    D = 0.17
    # Prandlt number
    Pr = 0.71

    vr = COMFA_vr(vw, va)
        
    Re =  vr * D / v

    if Re < 4000:
        A = 0.683
        n = 0.466
    elif ((Re >= 4000) and (Re < 40000)):
        A = 0.193
        n = 0.618
    else: # Re >= 40000
        A = 0.0266
        n = 0.805

    ra = 0.17 / (A * Re**n * Pr**0.33 * k)

    # A = Re
    # n = Re
    
    # L = Re < 4000
    # M = ((Re >= 4000) and (Re < 40000))
    # H = Re >= 40000
        
    # A[L] = 0.683  
    # n[L] = 0.466 

    # A[M] = 0.193
    # n[M] = 0.618
    
    # A[H] = 0.0266
    # n[H] = 0.805
    
    # Written by N. Kenny April 2006. Translated from MATLAB to Python by Nils Wallenberg 2022.

    return ra

def COMFA_Es(Mact, Ta, RH, age, kid):
    # Calculates sensible evaporative heat losses through Perspiration (W/m2)
    # with inputs Tair, Mact, RH

    MET = COMFA_MET(Mact, Ta, RH)

    if kid:
        Es = 0.42 * (MET - 58) * 0.5
    else:
        Es = 0.42 * (MET - 58)

    # Equation from Fanger (1970) for "comfortable" persons. 
    # Es = 0 if Met < 58.   

    # Written by Natasha Kenny, June 2006 for input into the COMFA energy
    # balance model under Evaporative Heat Loss (EVAP)
    # Now used by J. Vanos for Master's project, 2009. 

    # Based on data presented by Fanger (1972), Brown and
    # Gillespie (1986) suggest that body tissue resistance
    # decreases with Mact.

    return Es

def COMFA_rsk(Mact, Ta, RH, age, kid):
    # Same as rt (or skin tissue resistance) in papers. 
    # New equation added to COMFA model, from Kenny et al. (2009b in IJB).
    # Adapted April 2009 by Jenni Vanos for input into the COMFA equation under
    # Convective Heat Loss or Gain (CONV)

    # calculates the resistance (s/m) of the skin (body tissue) to heat transfer in
    # inputs: rho = density of air = ~1.16kg/m3; Cp = heat capacity of air =
    # 1000J/kg K), Es = sensible evaporative heat losses.

    # Evaporatve heat loss through perspiration. 
    Es = COMFA_Es(Mact, Ta, RH, age, kid) 
    # 1212 is volumetric heat capacity (or rho.*Cp)
    rsk = (1212) / (0.13 * Es + 15)

    # See Kenny et al. (2009b) pg 434 for derivation of rsk based on Kerslake 1972. 

    return rsk

def COMFA_Tc(Mact, Ta, RH):
    # Calculates core body temperature with inputs Mact (Metabolic Activity, W/m2),
    # Tair (air temperature in C), RH (relative humdity, %)

    MET = COMFA_MET(Mact, Ta, RH)

    Tc = 36.5 + (0.0043 * MET)

    # Written April 2009 by Jenni Vanos from Natasha Kenny, June 2006, for input into the COMFA energy balance 
    # model under Convective Heat Gain or Loss (CONV). See Kenny et al. (2009b)
    # in IJB for details. 

    # note this is not created for high intensity activity. 

    return Tc

def COMFA_Tsk(Mact, Ta, RH, vw, va, rco, age, kid):
    # Calculates skin temperature for input into the COMFA energy balance
    # equation with inputs Mact (metabolic activity, W/m2), Tair (air temp,
    # degrees C), rco (clothing resistance s/m), Vr (relative air
    # velocity m/s)

    # s m-1
    rsk = COMFA_rsk(Mact, Ta, RH, age, kid); 
    # s m-1
    ra = COMFA_ra(vw, va)
    # degC
    Tc = COMFA_Tc(Mact, Ta, RH)
    # s m-1
    rc = COMFA_rc(va, rco)
    # degC
    Tsk = (( Tc - Ta ) / (rsk + rc + ra)) * (ra + rc) + Ta

    # all resisteances are in s/m.
    
    # Equation for Tsk adapted from Kenny et al. (2009b) in IJB. Modified April 2009 by
    # Jenni Vanos. 

    return Tsk

def COMFA_CONV(Mact, Ta, RH, vw, va, rco, weight, height, age, kid):
    # Calculates the total heat loss (W/m2) by convection
    # with inputs Tair (air temperature, degrees C), Mact (metabolic activity,
    # W/m2), rco is the static clothing resistance (s/m),
    # vr (realtive wind velocity, m/s) (vr = vw when  va = 0) where va =
    # activity velocity, and vw = wind velocity

    # clothing resistance (Input should be va, not Ta as in original code?)
    rc = COMFA_rc(va, rco) 
    # resistance of boundary layer
    ra = COMFA_ra(vw, va)
    # Skin temperature
    Tsk = COMFA_Tsk(Mact, Ta, RH, vw, va, rco, age, kid) 
    # Final convection value 
    if kid:
        # Body to surface area of kid
        BSA_k = weight**0.5378 * height**0.3964 * 0.0242
        BMI_k = weight/(height/100)**2
        # Body-to-surface area of adult (weight = 65 kg and height = 176 cm, according to Cheng & Brown, 2020)
        adult_weight = 65
        adult_height = 176
        BSA_a = adult_weight**0.5378 * adult_height**0.3964 * 0.0242
        BMI_a = adult_weight/(adult_height/100)**2

        CONV = (1212 * (Tsk - Ta) / (rc + ra)) * ((BSA_k/weight)/(BSA_a/adult_weight))
    else:
        CONV = 1212 * (Tsk - Ta) / (rc + ra)

    # 1212 is just an  pre-calculated volumetric heat capapcity of air at 20oC ("rhoCp") of air (J m?3 K?1), 
    # assuming air density (rho) of ~1.2kg m-3 and Cp of 1000 J kg-1 K-1 at ~20oC,
    # which can be adapted based on air temperature 
    
    # Initally modified April, 2007 by Natasha Kenny to incorporate revisions of heat flow from surface
    # temp to air temp, rather than from core temp to air temp. See Kenny et al.
    # (2009a) IJB. 

    # Adapted Aug 2009 by Jenni Vanos in order to incorporate new Tsk
    # equation. Additional notes on rhoCp added by JV 2020.

    return CONV

def COMFA_rcv(vw, va, rcvo):
    # Written by Jenni Vanos April, 2009 for use in the updated COMFA model,
    # according to findings and changes made by Kenny et al. (2009). This is
    # equation 6 from Kenny et al. (2009b), and has inputs rcvo (or the static 
    # clothing vapour resistance, which is defined for standardized conditions 
    # (static body, wind still, i.e. speed < 0.2 m s-1) (ISO, 2007). When air 
    # movement is present, or when the body moves, this will affect the vapour 
    # resistance (typically lowering it), in which case it is referred to as the 
    # resultant or dynamic basic water vapour resistance (ISO, 2007).
    # rcvo can be calculated for input using the equation:

    # rcvo = Icl.*0.1555.*0.18.*18400; %ISO (2007) conversion from clo to s/m.

    # Where Icl is found in tables for specific clothing ensembles from ISO (2007), 
    # and 18400 is a conversion factor from Re,cl (vapour resistance, in m2kPaW-1) 
    # to rcvo, using Lv = 2.5×106 J kg-1, rho = 1.16 kg m-3, and Pa = 98kPa. 
    # See Kenny et al. (2009b) for further reference. 

    vr = COMFA_vr(vw, va)

    rcv = rcvo * (-0.80 * (1 - np.exp(-vr / 1.095))+1)

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    # Alternative method for standing individual when clothing isn't known: -
    # UTCI regression equation (Havenith 2012 IJB). 
    
    # Calculates the clothing resistance for input for the COMFA energy balance 
    # equation with inputs of Tair (degrees C) to calculate the Icl (clo)based 
    # on air temperature from the UTCI equation. See Vanos, 2012b. 
    # For use on standing individuals only. Can be slowly walking or standing.

    # Written Oct 2009 by Jenni Vanos
    # Because we are assuming little to no activity speed (va) then rc = rco.

    # Icl = 1.372 - (0.01866*Tair) - (0.0004849*(Tair.^2)) - (0.000009333*(Tair.^3)); %Icl here is in clo.
    # Icl equation is from UTCI, and more info can be found in Vanos et al.
    # (2012b). 
    # rcvo = Icl.*0.1555.*0.18.*18400; %ISO (2007) conversion from clo to s/m.

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    return rcv

def COMFA_rav(vw, va):    
    # Calculates the resistance of the boundary air layer to vapour transport,
    # with input of relative wind velocity (m/s)

    ra = COMFA_ra(vw, va)
    # (s/m)
    rav = 0.92 * ra

    # Adapated April 2009 by Jenni Vanos to incorporate vr (m/s), which includes va and vw (m/s).
    # Units of rav are in s/m. 

    return rav

def COMFA_qsk(Mact, Ta, RH, vw, va, rco, age, kid):
    # Calculates the saturation specific humidity (kg/kg) at skin temperature for input into the
    # COMFA energy balance equation with inputs Mact (metabolic activity), Tair
    # (degrees C), RH (%), rco, va and Vr. 
    # Assuming atmospheric pressure (Pa) is 1013 mb (101.3kPa)

    Tsk = COMFA_Tsk(Mact, Ta, RH, vw, va, rco, age, kid)
    
    esk = 0.61365 * np.exp((17.502 * Tsk) / (240.97 + Tsk))

    qsk = 0.622 * (esk / (101.3 - esk))

    return qsk

def COMFA_qa(Ta, RH):
    # Calculates saturation specific humidity (kg/kg) at air temperatre for input
    # into the COMFA energy balance equation, with inputs Tair (air
    # temperature, degrees C), RH (relative humidity, %).
    # assuming atmospheric pressure is 1013 mb (101.3kPa)

    e = COMFA_e(Ta, RH)
    
    qa = 0.622 * e / (101.3 - e)
    
    # svp = Tetens_svp(Tair)
    #  
    # qa = svp.*(RH./100)
    
    # Written July, 2006 by Natasha Kenny, modified to reflect the changes to
    # this equation. See usage also in Kenny et al. (2009b) 
    # Re-written April 2009 by Jenni Vanos. 

    return qa

def COMFA_Ei(Mact, Ta, RH, vw, va, rco, rcvo, age, kid):
    # Calculates insensible evaporative heat losses in the COMFA energy balance
    # equation with inputs Tair (air temp, degrees C), RH (relative humidity %)
    # vw (wind velocity (m/s), va (acivity velocity), Mact (Metabolic activity, W/m2), 
    # rco(clothing resistance, s/m), rcvo (clothing vapour resistance s/m)

    # Density of air in kg/m3
    rho = 1.16
    # Latent heat of vapourization (J/kg)
    Lv = 2.442e+6
    # Resistance of skin to vapour transfer (s/m) - Campbell 1977
    rskv = 7.7e+3
    # clothing vapour resistance (s/m)
    rcv = COMFA_rcv(vw, va, rcvo)
    # Resistance of boundary air layer to vapour transport
    rav = COMFA_rav(vw, va)
    # update
    qsk = COMFA_qsk(Mact, Ta, RH, vw, va, rco, age, kid)
    # update
    qa = COMFA_qa(Ta, RH) 

    Ei = rho * Lv * ((qsk - qa) / (rcv + rav + rskv))
    
    # note Ei is very unimportant.
    
    # Written July, 2006 by Natasha Kenny, adapted April 2009 by Jenni
    # Vanos to incorporate changes with vr and rcv from changes made by
    # Kenny et al. (2009). 

    return Ei

def COMFA_Etot(Mact, Ta, RH, vw, va, rco, rcvo, age, kid):
    # Calculates the total evaporative heat loss for input into the COMFA
    # equation with inputs (Tair air temp in C, RH Relative Humidity %, vw, 
    # wind velocity (m/s), Mact, metabolic activity W/m2, rco (clothing resistance, s/m), va activity velocity (m/s), rcvo (clothing vapour resistance, s/m) not to be confused with
    # EVAP which is the final input in the equation

    Ei = COMFA_Ei(Mact, Ta, RH, vw, va, rco, rcvo, age, kid)

    Es = COMFA_Es(Mact, Ta, RH, age, kid)

    Etot = Ei + Es
    
    # Written July, 2006 by Natasha Kenny. Adapted April 2009 by Jenni Vanos to
    # include vr, and subtract P based on Kenny et al. (2009). 

    return Etot

def COMFA_Em(Mact, Ta, RH, vw, va, rco, rcvo, age, kid):
    # Calculates evaporative maximum, Em (W/m2) (where Em = the maximum possible evapoaration 
    # possible given the environment and clothing, and assumes a skin wettedness value (w) of 1)
    # for input into COMFA equation.
     
    # Inputs vr (relative wind velocity, m/s) (vr = vw when va=0), Mact (metabolic
    # activity), Tair (air temperature, degrees C), RH (relative humdity, %)
    
    # Density of air kg m-3
    rho = 1.16
    # latent heat of vaporization J kg-1
    Lv = 2.442e+6
    
    rav = COMFA_rav(vw, va)
    
    rcv = COMFA_rcv(vw, va, rcvo)
    
    qsk = COMFA_qsk(Mact, Ta, RH, vw, va, rco, age, kid)
    
    qa = COMFA_qa(Ta, RH)
       
    Em = rho * Lv * ((qsk - qa) / (rcv + rav))

    # Written by Natasha Kenny, June 2006 for input into the COMFA equation.
    # Adapted by Jenni Vanos April 2009 for addition of changes by Kenny et al.
    # (2009). (-P and + vr, va, vw)

    return Em

def COMFA_EVAP(Mact, Ta, RH, vw, va, rco, rcvo, age, kid):

    # Calculates the total evaporative heat losses for input into the COMFA
    # energy balance model with inputs Ei (insensible heat loss through the skin
    # and Es (sensible heat losses through the skin) W/m2

    Etot = COMFA_Etot(Mact, Ta, RH, vw, va, rco, rcvo, age, kid)

    Em = COMFA_Em(Mact, Ta, RH, vw, va, rco, rcvo, age, kid)
 
    if Etot <= Em:
        EVAP = Etot
    else:
        EVAP = Em

    # EVAP = np.zeros((Em.shape[0]))
    
    # ind = Etot <= Em
 
    # EVAP[ind] = Etot[ind]

    # The lower of E or Em is used in the energy budget equation.
    
    # Initially written by Kenny, modified by Jenni Vanos, April 2009. 

    return EVAP

def COMFA_Ts(Mact, Ta, RH, vw, va, rco, age, kid):
    # Calculates the average surface temperature of a clothed person (degrees C) with inputs Tair (air
    # temperature, degrees C), rco (clothing resistance,  s/m), vw (air velocity, s/m), Mact (Metabolic Activity, W/m2)

    rc = COMFA_rc(va, rco)
    # Resistance of boundary layer (s/m)
    ra = COMFA_ra(vw, va)
    # skin temperature (degC)
    Tsk = COMFA_Tsk(Mact, Ta, RH, vw, va, rco, age, kid)
    # Average surface temperature of a person
    Ts = ((Tsk - Ta) / (rc + ra)) * ra + Ta

    # Adapted from Natasha on April 2009 to include vr, va and take out P. 
    # Modified April, 2007 by Natasha Kenny to revise the heat flow from the skin surface to the
    # air rather than from the core temperature to the air.

    return Ts

def COMFA_TREMIT(Mact, Ta, RH, vw, va, rco, age, kid):
    # Calculates the terrestrial radiation losses from the body for input into the COMFA
    # energy balance equation with inputs of Tair (air temp, degrees C), rco (clothing
    # insulation (s/m)), wind velocity (m/s), activity velocity (m/s)), and Mact,
    # (metabolic activity (w/m2)).
    
    # surface temperature (C)
    Ts = COMFA_Ts(Mact, Ta, RH, vw, va, rco, age, kid)

    # 0.95 is the emissivity of human (Campbell & Norman 1998)
    # 0.78 is the radiative surface area of a human standing (Kerslake, 1972)
    # 0.7 for sitting/on bicycle
    TREMIT = 0.78 * (0.95 * 5.67e-8 * (Ts + 273.15)**4) 

    # Adapted April 2009 by J. Vanos to incorporate new vr geomtric equation (vr =
    # COMFA_vr(va, vw)), and take out P, which is no longer needed in the COMFA model
    # (Kenny et al., 2009a,b). 

    # modified April, 2007 by Natasha Kenny to incorportate emissivity of 0.95 and radiative
    # surface of a human 0.78 (this is for when walking or running). 

    return TREMIT

def COMFA_BUDGET(Mact, Ta, RH, vw, va, rco, rcvo, weight, height, age, kid):
    # Calculates the four energy fluxes MET, CONV, EVAP, and TREMIT in the COMFA energy budget, with inputs 
    # Tair (air temperature, degrees C), Metabolic activity (W/m2), Relative Humidty (%),
    # Wind Velocity (m/s), activity velocity (va, m s-1), static clothing resistance (rco, sm-1), static clothing vapor reistance (rcvo, s m-1). 
    # Note can also make Rabs an input. See below.

    ### This COFMA model is general for people at or near comfort, and is not
    ### tested yet on extreme conditions. 

    # %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% 
    # %%%% CLOTHING - generally can estimate Icl value (clo) and convert them to
    # %%%% resistances as follows: 

    # Rco: rco = Icl * 186.6. Or convert from conductivity based on 1 Clo = 0.1555 (m 2K W-1).		

    # NOTE: Icl values can be found in clo or conductivity values from ISO 9920 2007. 

    # Rcvo: Can convert from a conductivity value of Icl, where Icl is found in tables for specific clothing ensembles from ISO (2007). 
    # 1) convert from Icl to Re,cl (m2 kPa W-1):  Re,cl = Icl (m 2K W-1)*0.18 (constant from ISO 9920 pg 12 based on 1 or 2-layer clothing ensembles). 
    # 2) Convert Re,cl to rcvo:  rcvo = Re,cl*18,400, where 18400 is a conversion factor from Re,cl (vapour resistance, in m2kPaW-1) 		
    # to rcvo, using Lv = 2.5*106 J kg-1, rho = 1.16 kg m-3, and Pa = 98kPa. 		
    # %%%%%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% %%%%%%%%%%% 
    
    # W m-2 %%%%random value ... Here you need to link to another dataset that calcs Rabs or bring o link to more doce. 
    # Rabs = 400
    
    MET = COMFA_MET(Mact, Ta, RH)

    CONV = COMFA_CONV(Mact, Ta, RH, vw, va, rco, weight, height, age, kid)

    EVAP = COMFA_EVAP(Mact, Ta, RH, vw, va, rco, rcvo, age, kid)
    
    TREMIT = COMFA_TREMIT(Mact, Ta, RH, vw, va, rco, age, kid)
    
    # BUDGET = MET + Rabs - CONV - EVAP - TREMIT

    # all are in W m-2

    # return BUDGET

    return MET, CONV, EVAP, TREMIT


from builtins import range

# -*- coding: utf-8 -*-
__author__ = "xlinfr"

import math

import numpy as np
import scipy.ndimage.interpolation as sc
from tqdm import tqdm


def findwalls(a, walllimit):
    # This function identifies walls based on a DSM and a wall-height limit
    # Walls are represented by outer pixels within building footprints
    #
    # Fredrik Lindberg, Goteborg Urban Climate Group
    # fredrikl@gvc.gu.se
    # 20150625

    col = a.shape[0]
    row = a.shape[1]
    walls = np.zeros((col, row))
    domain = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    index = 0
    for i in tqdm(np.arange(1, row - 1)):
        for j in np.arange(1, col - 1):
            dom = a[j - 1 : j + 2, i - 1 : i + 2]
            walls[j, i] = np.max(dom[np.where(domain == 1)])  # new 20171006
            index = index + 1

    walls = np.copy(walls - a)  # new 20171006
    walls[(walls < walllimit)] = 0

    walls[0 : walls.shape[0], 0] = 0
    walls[0 : walls.shape[0], walls.shape[1] - 1] = 0
    walls[0, 0 : walls.shape[0]] = 0
    walls[walls.shape[0] - 1, 0 : walls.shape[1]] = 0

    return walls


def filter1Goodwin_as_aspect_v3(walls, scale, a):
    """
    tThis function applies the filter processing presented in Goodwin et al (2010) but instead for removing
    linear fetures it calculates wall aspect based on a wall pixels grid, a dsm (a) and a scale factor

    Fredrik Lindberg, 2012-02-14
    fredrikl@gvc.gu.se

    Translated: 2015-09-15

    :param walls:
    :param scale:
    :param a:
    :return: dirwalls
    """

    row = a.shape[0]
    col = a.shape[1]

    filtersize = np.floor((scale + 0.0000000001) * 9)
    if filtersize <= 2:
        filtersize = 3
    else:
        if filtersize != 9:
            if filtersize % 2 == 0:
                filtersize = filtersize + 1

    filthalveceil = int(np.ceil(filtersize / 2.0))
    filthalvefloor = int(np.floor(filtersize / 2.0))

    filtmatrix = np.zeros((int(filtersize), int(filtersize)))
    buildfilt = np.zeros((int(filtersize), int(filtersize)))

    filtmatrix[:, filthalveceil - 1] = 1
    n = filtmatrix.shape[0] - 1
    buildfilt[filthalveceil - 1, 0:filthalvefloor] = 1
    buildfilt[filthalveceil - 1, filthalveceil : int(filtersize)] = 2

    y = np.zeros((row, col))  # final direction
    z = np.zeros((row, col))  # temporary direction
    x = np.zeros((row, col))  # building side
    walls[walls > 0.5] = 1

    for h in tqdm(range(0, 180)):  # =0:1:180 #%increased resolution to 1 deg 20140911
        filtmatrix1temp = sc.rotate(filtmatrix, h, order=1, reshape=False, mode="nearest")  # bilinear
        filtmatrix1 = np.round(filtmatrix1temp)
        # filtmatrix1temp = sc.imrotate(filtmatrix, h, 'bilinear')
        # filtmatrix1 = np.round(filtmatrix1temp / 255.)
        # filtmatrixbuildtemp = sc.imrotate(buildfilt, h, 'nearest')
        filtmatrixbuildtemp = sc.rotate(buildfilt, h, order=0, reshape=False, mode="nearest")  # Nearest neighbor
        # filtmatrixbuild = np.round(filtmatrixbuildtemp / 127.)
        filtmatrixbuild = np.round(filtmatrixbuildtemp)
        index = 270 - h
        if h == 150:
            filtmatrixbuild[:, n] = 0
        if h == 30:
            filtmatrixbuild[:, n] = 0
        if index == 225:
            # n = filtmatrix.shape[0] - 1  # length(filtmatrix);
            filtmatrix1[0, 0] = 1
            filtmatrix1[n, n] = 1
        if index == 135:
            # n = filtmatrix.shape[0] - 1  # length(filtmatrix);
            filtmatrix1[0, n] = 1
            filtmatrix1[n, 0] = 1

        for i in range(int(filthalveceil) - 1, row - int(filthalveceil) - 1):  # i=filthalveceil:sizey-filthalveceil
            for j in range(
                int(filthalveceil) - 1, col - int(filthalveceil) - 1
            ):  # (j=filthalveceil:sizex-filthalveceil
                if walls[i, j] == 1:
                    wallscut = (
                        walls[
                            i - filthalvefloor : i + filthalvefloor + 1,
                            j - filthalvefloor : j + filthalvefloor + 1,
                        ]
                        * filtmatrix1
                    )
                    dsmcut = a[
                        i - filthalvefloor : i + filthalvefloor + 1,
                        j - filthalvefloor : j + filthalvefloor + 1,
                    ]
                    if z[i, j] < wallscut.sum():  # sum(sum(wallscut))
                        z[i, j] = wallscut.sum()  # sum(sum(wallscut));
                        if np.sum(dsmcut[filtmatrixbuild == 1]) > np.sum(dsmcut[filtmatrixbuild == 2]):
                            x[i, j] = 1
                        else:
                            x[i, j] = 2

                        y[i, j] = index

    y[(x == 1)] = y[(x == 1)] - 180
    y[(y < 0)] = y[(y < 0)] + 360

    grad, asp = get_ders(a, scale)

    y = y + ((walls == 1) * 1) * ((y == 0) * 1) * (asp / (math.pi / 180.0))

    dirwalls = y

    return dirwalls


def cart2pol(x, y, units="deg"):
    radius = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    if units in ["deg", "degs"]:
        theta = theta * 180 / np.pi
    return theta, radius


def get_ders(dsm, scale):
    # dem,_,_=read_dem_grid(dem_file)
    dx = 1 / scale
    # dx=0.5
    fy, fx = np.gradient(dsm, dx, dx)
    asp, grad = cart2pol(fy, fx, "rad")
    grad = np.arctan(grad)
    asp = asp * -1
    asp = asp + (asp < 0) * (np.pi * 2)
    return grad, asp



import numpy as np
from copy import deepcopy
from . import sunlit_shaded_patches

''' This function defines if a patch seen from a pixel is sky, building or vegetation. 
    It also calculates if a building patch is sunlit or shaded. From this it estimates 
    corresponding longwave radiation originating from each surface.'''

def define_patch_characteristics(solar_altitude, solar_azimuth,
                             patch_altitude, patch_azimuth, steradian,
                             asvf,
                             shmat, vegshmat, vbshvegshmat,
                             Lsky_down, Lsky_side, Lsky, Lup,
                             Ta, Tgwall, ewall,
                             rows, cols):
    
    # Stefan-Boltzmann's Constant
    SBC = 5.67051e-8

    # Degrees to radians
    deg2rad = np.pi / 180

    # Define variables
    Ldown = np.zeros((rows,cols))
    Ldown_sky = np.zeros((rows,cols))
    Ldown_veg = np.zeros((rows,cols))
    Ldown_sun = np.zeros((rows,cols))
    Ldown_sh = np.zeros((rows,cols))
    Ldown_ref = np.zeros((rows,cols))

    Lside = np.zeros((rows,cols))
    Lside_sky = np.zeros((rows,cols))
    Lside_veg = np.zeros((rows,cols))
    Lside_sun = np.zeros((rows,cols))
    Lside_sh = np.zeros((rows,cols))
    Lside_ref = np.zeros((rows,cols))

    Least = np.zeros((rows,cols))
    Lwest = np.zeros((rows,cols))
    Lnorth = np.zeros((rows,cols))
    Lsouth = np.zeros((rows,cols))

    # Define patch characteristics (sky, vegetation or building, and sunlit or shaded if building)
    for idx in range(patch_altitude.shape[0]):
        # Calculations for patches on sky, shmat = 1 = sky is visible
        temp_sky = ((shmat[:,:,idx] == 1) & (vegshmat[:,:,idx] == 1))
        
        # Longwave radiation from sky to vertical surface
        Ldown_sky += temp_sky * Lsky_down[idx,2]
        
        # Longwave radiation from sky to horizontal surface
        Lside_sky += temp_sky * Lsky_side[idx,2]

        # Calculations for patches that are vegetation, vegshmat = 0 = shade from vegetation
        temp_vegsh = ((vegshmat[:,:,idx] == 0) | (vbshvegshmat[:,:,idx] == 0))
        # Longwave radiation from vegetation surface (considered vertical)
        vegetation_surface = ((ewall * SBC * ((Ta + 273.15) ** 4)) / np.pi)

        # Longwave radiation reaching a vertical surface
        Lside_veg += vegetation_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh
        
        # Longwave radiation reaching a horizontal surface
        Ldown_veg += vegetation_surface * steradian[idx] * np.sin(patch_altitude[idx] * deg2rad) * temp_vegsh
        
        # Portion into cardinal directions to be used for standing box or POI output
        if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
            Least += temp_sky * Lsky_side[idx,2] * np.cos((90 - patch_azimuth[idx]) * deg2rad)
            Least += vegetation_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((90 - patch_azimuth[idx]) * deg2rad)
        if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
            Lsouth += temp_sky * Lsky_side[idx,2] * np.cos((180 - patch_azimuth[idx]) * deg2rad)
            Lsouth += vegetation_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((180 - patch_azimuth[idx]) * deg2rad)
        if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
            Lwest += temp_sky * Lsky_side[idx,2] * np.cos((270 - patch_azimuth[idx]) * deg2rad)
            Lwest += vegetation_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((270 - patch_azimuth[idx]) * deg2rad)
        if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
            Lnorth += temp_sky * Lsky_side[idx,2] * np.cos((0 - patch_azimuth[idx]) * deg2rad)
            Lnorth += vegetation_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_vegsh * np.cos((0 - patch_azimuth[idx]) * deg2rad)

        # Calculations for patches that are buildings, shmat = 0 = shade from buildings
        temp_vbsh = (1 - shmat[:,:,idx]) * vbshvegshmat[:,:,idx]
        temp_sh = (temp_vbsh == 1)
        azimuth_difference = np.abs(solar_azimuth - patch_azimuth[idx])

        # Longwave radiation from sunlit surfaces
        sunlit_surface = ((ewall * SBC * ((Ta + Tgwall + 273.15) ** 4)) / np.pi)
        # Longwave radiation from shaded surfaces
        shaded_surface = ((ewall * SBC * ((Ta + 273.15) ** 4)) / np.pi)
        if ((azimuth_difference > 90) and (azimuth_difference < 270) and (solar_altitude > 0)):
            # Calculate which patches defined as buildings that are sunlit or shaded
            sunlit_patches, shaded_patches = sunlit_shaded_patches.shaded_or_sunlit(solar_altitude, solar_azimuth, patch_altitude[idx], patch_azimuth[idx], asvf)
            
            # Calculate longwave radiation from sunlit walls to vertical surface
            Lside_sun += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh
            # Calculate longwave radiation from shaded walls to vertical surface
            Lside_sh += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh
            
            # Calculate longwave radiation from sunlit walls to horizontal surface
            Ldown_sun += sunlit_surface * sunlit_patches * steradian[idx] * np.sin(patch_altitude[idx] * deg2rad) * temp_sh
            # Calculate longwave radiation from shaded walls to horizontal surface
            Ldown_sh += shaded_surface * shaded_patches * steradian[idx] * np.sin(patch_altitude[idx] * deg2rad) * temp_sh
            
            # Portion into cardinal directions to be used for standing box or POI output
            if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
                Least += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx]) * deg2rad)
                Least += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx]) * deg2rad)
            if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
                Lsouth += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx]) * deg2rad)
                Lsouth += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx]) * deg2rad)
            if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
                Lwest += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx]) * deg2rad)
                Lwest += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx]) * deg2rad)
            if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
                Lnorth += sunlit_surface * sunlit_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx]) * deg2rad)
                Lnorth += shaded_surface * shaded_patches * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx]) * deg2rad)

        else:
            # Calculate longwave radiation from shaded walls reaching a vertical surface
            Lside_sh += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh
            
            # Calculate longwave radiation from shaded walls reaching a horizontal surface
            Ldown_sh += shaded_surface * steradian[idx] * np.sin(patch_altitude[idx] * deg2rad) * temp_sh

            # Portion into cardinal directions to be used for standing box or POI output
            if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
                Least += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx]) * deg2rad)
            if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
                Lsouth += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx]) * deg2rad)
            if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
                Lwest += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx]) * deg2rad)
            if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
                Lnorth += shaded_surface * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx]) * deg2rad)

    # Calculate reflected longwave in each patch
    reflected_on_surfaces = (((Ldown_sky+Lup)*(1-ewall)*0.5) / np.pi)
    for idx in range(patch_altitude.shape[0]):
        temp_sh = ((shmat[:,:,idx] == 0) | (vegshmat[:,:,idx] == 0) | (vbshvegshmat[:,:,idx] == 0))
        
        # Reflected longwave radiation reaching vertical surfaces
        Lside_ref += reflected_on_surfaces * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh
        
        # Reflected longwave radiation reaching horizontal surfaces
        Ldown_ref += reflected_on_surfaces * steradian[idx] * np.sin(patch_altitude[idx] * deg2rad) * temp_sh
        
        # Portion into cardinal directions to be used for standing box or POI output
        if (patch_azimuth[idx] > 360) or (patch_azimuth[idx] < 180):
            Least += reflected_on_surfaces * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((90 - patch_azimuth[idx]) * deg2rad)
        if (patch_azimuth[idx] > 90) and (patch_azimuth[idx] < 270):
            Lsouth += reflected_on_surfaces * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((180 - patch_azimuth[idx]) * deg2rad)
        if (patch_azimuth[idx] > 180) and (patch_azimuth[idx] < 360):
            Lwest += reflected_on_surfaces * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((270 - patch_azimuth[idx]) * deg2rad)
        if (patch_azimuth[idx] > 270) or (patch_azimuth[idx] < 90):
            Lnorth += reflected_on_surfaces * steradian[idx] * np.cos(patch_altitude[idx] * deg2rad) * temp_sh * np.cos((0 - patch_azimuth[idx]) * deg2rad)

    # Sum of all Lside components (sky, vegetation, sunlit and shaded buildings, reflected)
    Lside = Lside_sky + Lside_veg + Lside_sh + Lside_sun + Lside_ref

    # Sum of all Lside components (sky, vegetation, sunlit and shaded buildings, reflected)
    Ldown = Ldown_sky + Ldown_veg + Ldown_sh + Ldown_sun + Ldown_ref

    return Ldown, Lside, Lside_sky, Lside_veg, Lside_sh, Lside_sun, Lside_ref, Least, Lwest, Lnorth, Lsouth


''' This function creates a hemispheric image based on patch characteristics '''
def hemispheric_image(poi, sh, vegsh, vbshvegsh, voxelmat, wallScheme):
    patch_characteristics = np.zeros((sh.shape[2], poi.shape[0]))
    for idx in range(poi.shape[0]):
        for idy in range(sh.shape[2]):
            # Calculations for patches on sky, shmat = 1 = sky is visible
            temp_sky = ((sh[:,:,idy] == 1) & (vegsh[:,:,idy] == 1))
            # Calculations for patches that are vegetation, vegshmat = 0 = shade from vegetation
            temp_vegsh = ((vegsh[:,:,idy] == 0) | (vbshvegsh[:,:,idy] == 0))
            # Calculations for patches that are buildings, shmat = 0 = shade from buildings
            temp_vbsh = (1 - sh[:,:,idy]) * vbshvegsh[:,:,idy]
            temp_sh = (temp_vbsh == 1)
            if wallScheme:
                temp_sh_w = temp_sh * voxelmat[:, :, idy]
                temp_sh_roof = temp_sh * (voxelmat[:, :, idy] == 0)
            else:
                temp_sh_w = 0
                temp_sh_roof = 0
            # Sky patch
            if temp_sky[int(poi[idx, 2]), int(poi[idx, 1])]:
                patch_characteristics[idy,idx] = 1.8
            # Vegetation patch
            elif (temp_vegsh[int(poi[idx, 2]), int(poi[idx, 1])]):
                patch_characteristics[idy,idx] = 2.5
            # Building patch
            elif (temp_sh[int(poi[idx, 2]), int(poi[idx, 1])]):
                if wallScheme:
                    if (temp_sh_w[int(poi[idx, 2]), int(poi[idx, 1])]):
                        patch_characteristics[idy, idx] = 4.5
                    elif (temp_sh_roof[int(poi[idx, 2]), int(poi[idx, 1])]):
                        # patch_characteristics[idy, idx] = 6.0
                        patch_characteristics[idy, idx] = 4.5
                else:
                    patch_characteristics[idy,idx] = 4.5
            # Roof patch
            #elif
    return patch_characteristics



import xarray as xr
import rioxarray
import numpy as np

def walls_as_netcdf(voxelTable, rows, cols, timeSlots, iteration, dsm, raster_path, output_path):
    '''This function creates a 4D NetCDF with wall temperatures and corresponding emitted longwave radiation'''
    # rows = number of rows (latitudinal position)
    # cols = number of columns (longitudinal position)
    # level = number of voxel levels (elevation position)
    # raster_path is used to load an existing .tif layer and create arrays with latitudes and longitudes to be used in xarray/NetCDF

    # Highest number of voxels used to determine z/height level of NetCDF
    # levels = voxelTable.loc[voxelTable['voxelHeight'] == voxelTable['voxelHeight'].max(), 'voxelHeight'].to_numpy()[0].astype(int)

    levels = voxelTable.loc[voxelTable['voxelHeightMasl'] == voxelTable['voxelHeightMasl'].max(), 'voxelHeightMasl'].to_numpy()[0].astype(int)

    # Range of height levels
    height_levels = np.arange(1, levels+1)

    # Create empty numpy array to fill with wall temperatures from current time step
    wallTemperature = np.full((cols, rows, levels), np.nan, dtype=np.float32)

    # Add current time step wall temperature and longwave radiation to numpy array, which will be used to update the NetCDF.
    #for y, x, z, wallTemp in zip(voxelTable['ypos'].astype(int), voxelTable['xpos'].astype(int), voxelTable['voxelHeight'].astype(int), voxelTable['wallTemperature'].astype(np.float32)):
    for y, x, z, wallTemp in zip(voxelTable['ypos'].astype(int), voxelTable['xpos'].astype(int), voxelTable['voxelHeightMasl'].astype(int), voxelTable['wallTemperature'].astype(np.float32)):
        wallTemperature[x, y, z-1] = wallTemp

    # NetCDF compression
    comp = dict(zlib=True, 
                complevel=5,
                chunksizes=(100, 100, 10, 1))

    # If first time step, create an empty NetCDF to fill with values
    if iteration == 0:
        # raster_file = xr.open_rasterio(raster_path)
        # raster_file = xr.open_dataset(raster_path, decode_coords='all', engine='rasterio')
        raster_file = rioxarray.open_rasterio(raster_path)
        lat = raster_file.y.to_numpy()
        lon = raster_file.x.to_numpy()
        # temp_data = np.zeros((cols, rows, levels, timeSlots.shape[0]))
        temp_data = np.full((cols, rows, levels, timeSlots.shape[0]), np.nan, dtype=np.float32)
        data_xr = xr.Dataset(
            data_vars=dict(
            wall_temperature=(["lon", "lat", "height", "time"], temp_data),
                ),
            coords=dict(
            lon=lon,
            lat=lat,
            height=height_levels,
            time=timeSlots
            ),
            attrs={"crs": raster_file.rio.crs.to_string()}
        )
        
        # Update wall temperature and longwave radiation for current timestep (iteration)
        data_xr.wall_temperature[:, :, :, iteration] = wallTemperature
        
        encodings = {var: comp for var in data_xr.data_vars}

        # Save as NetCDF
        data_xr.to_netcdf(output_path, engine="netcdf4", encoding=encodings)

        data_xr.close()
    # If not first time step, load existing NetCDF as an xarray dataset
    else:
        with xr.open_dataset(output_path, mode='r+') as data_xr:
            # Update wall temperature for current timestep (iteration)
            data_xr.wall_temperature[:, :, :, iteration] = wallTemperature
            # Save changes
            data_xr.to_netcdf(output_path, engine="netcdf4", mode="a")    


from __future__ import absolute_import
import numpy as np
from .sunonsurface_2015a import sunonsurface_2015a

def gvf_2015a(wallsun,walls,buildings,scale,shadow,first,second,dirwalls,Tg,Tgwall,Ta,emis_grid,ewall,alb_grid,SBC,albedo_b,rows, cols,Twater,lc_grid,landcover):
    
    azimuthA=np.arange(5, 359, 20) #Search directions for Ground View Factors (GVF)
    
    #### Ground View Factors ####
    gvfLup=np.zeros((rows,cols))
    gvfalb=np.zeros((rows,cols))
    gvfalbnosh=np.zeros((rows,cols))
    gvfLupE=np.zeros((rows,cols))
    gvfLupS=np.zeros((rows,cols))
    gvfLupW=np.zeros((rows,cols))
    gvfLupN=np.zeros((rows,cols))
    gvfalbE=np.zeros((rows,cols))
    gvfalbS=np.zeros((rows,cols))
    gvfalbW=np.zeros((rows,cols))
    gvfalbN=np.zeros((rows,cols))
    gvfalbnoshE=np.zeros((rows,cols))
    gvfalbnoshS=np.zeros((rows,cols))
    gvfalbnoshW=np.zeros((rows,cols))
    gvfalbnoshN=np.zeros((rows,cols))

    #  sunwall=wallinsun_2015a(buildings,azimuth(i),shadow,psi(i),dirwalls,walls);
    sunwall = (wallsun/walls*buildings) == 1  # new as from 2015a

    for j in np.arange(0, azimuthA.__len__()):
        _, gvfLupi, gvfalbi, gvfalbnoshi = sunonsurface_2015a(azimuthA[j], scale, buildings, shadow, sunwall, first,
                                                              second, dirwalls*np.pi/180, walls, Tg, Tgwall, Ta,
                                                              emis_grid, ewall, alb_grid, SBC, albedo_b, Twater,
                                                              lc_grid, landcover)
        
        gvfLup=gvfLup+gvfLupi
        gvfalb=gvfalb+gvfalbi
        gvfalbnosh=gvfalbnosh+gvfalbnoshi
        
        if (azimuthA[j] >= 0) and (azimuthA[j] < 180):
            gvfLupE=gvfLupE+gvfLupi
            gvfalbE=gvfalbE+gvfalbi
            gvfalbnoshE=gvfalbnoshE+gvfalbnoshi

        if (azimuthA[j] >= 90) and (azimuthA[j] < 270):
            gvfLupS=gvfLupS+gvfLupi
            gvfalbS=gvfalbS+gvfalbi
            gvfalbnoshS=gvfalbnoshS+gvfalbnoshi

        if (azimuthA[j] >= 180) and (azimuthA[j] < 360):
            gvfLupW=gvfLupW+gvfLupi
            gvfalbW=gvfalbW+gvfalbi
            gvfalbnoshW=gvfalbnoshW+gvfalbnoshi

        if (azimuthA[j] >= 270) or (azimuthA[j] < 90):
            gvfLupN=gvfLupN+gvfLupi
            gvfalbN=gvfalbN+gvfalbi
            gvfalbnoshN=gvfalbnoshN+gvfalbnoshi

    gvfLup=gvfLup/azimuthA.__len__()+SBC*emis_grid*(Ta+273.15)**4
    gvfalb=gvfalb/azimuthA.__len__()
    gvfalbnosh=gvfalbnosh/azimuthA.__len__()
    
    gvfLupE=gvfLupE/(azimuthA.__len__()/2)+SBC*emis_grid*(Ta+273.15)**4
    gvfLupS=gvfLupS/(azimuthA.__len__()/2)+SBC*emis_grid*(Ta+273.15)**4
    gvfLupW=gvfLupW/(azimuthA.__len__()/2)+SBC*emis_grid*(Ta+273.15)**4
    gvfLupN=gvfLupN/(azimuthA.__len__()/2)+SBC*emis_grid*(Ta+273.15)**4
    
    gvfalbE=gvfalbE/(azimuthA.__len__()/2)
    gvfalbS=gvfalbS/(azimuthA.__len__()/2)
    gvfalbW=gvfalbW/(azimuthA.__len__()/2)
    gvfalbN=gvfalbN/(azimuthA.__len__()/2)
    
    gvfalbnoshE=gvfalbnoshE/(azimuthA.__len__()/2)
    gvfalbnoshS=gvfalbnoshS/(azimuthA.__len__()/2)
    gvfalbnoshW=gvfalbnoshW/(azimuthA.__len__()/2)
    gvfalbnoshN=gvfalbnoshN/(azimuthA.__len__()/2)
    
    return gvfLup,gvfalb,gvfalbnosh,gvfLupE,gvfalbE,gvfalbnoshE,gvfLupS,gvfalbS,gvfalbnoshS,gvfLupW,gvfalbW,gvfalbnoshW,gvfLupN,gvfalbN,gvfalbnoshN


from pathlib import Path

import numpy as np

try:
    import pyproj
    import rasterio
    from rasterio.features import rasterize
    from rasterio.mask import mask
    from rasterio.transform import Affine, from_origin
    from shapely import geometry

    GDAL_ENV = False

except:
    from osgeo import gdal

    GDAL_ENV = True


def rasterise_gdf(gdf, geom_col, ht_col, bbox=None, pixel_size: int = 1):
    # Define raster parameters
    if bbox is not None:
        # Unpack bbox values
        minx, miny, maxx, maxy = bbox
    else:
        # Use the total bounds of the GeoDataFrame
        minx, miny, maxx, maxy = gdf.total_bounds
    width = int((maxx - minx) / pixel_size)
    height = int((maxy - miny) / pixel_size)
    transform = from_origin(minx, maxy, pixel_size, pixel_size)
    # Create a blank array for the raster
    raster = np.zeros((height, width), dtype=np.float32)
    # Burn geometries into the raster
    shapes = ((geom, value) for geom, value in zip(gdf[geom_col], gdf[ht_col], strict=True))
    raster = rasterize(shapes, out_shape=raster.shape, transform=transform, fill=0, dtype=np.float32)

    return raster, transform


def check_path(path_str: str | Path, make_dir: bool = False) -> Path:
    # Ensure path exists
    path = Path(path_str).absolute()
    if not path.exists() and not path.suffix:
        if make_dir:
            path.mkdir(parents=True, exist_ok=True)
        else:
            raise OSError(f"Path {path} does not exist.")
    return path


def save_raster(
    out_path_str: str, data_arr: np.ndarray, trf_arr: list[float], crs_wkt: str, no_data_val: float = -9999
):
    # Save raster using GDAL or rasterio
    out_path = check_path(out_path_str, make_dir=True)
    height, width = data_arr.shape
    if GDAL_ENV is False:
        trf = Affine.from_gdal(*trf_arr)
        crs = pyproj.CRS.from_wkt(crs_wkt)
        with rasterio.open(
            out_path,
            "w",
            driver="GTiff",
            height=height,
            width=width,
            count=1,
            dtype=data_arr.dtype,
            crs=crs,
            transform=trf,
            nodata=no_data_val,
        ) as dst:
            dst.write(data_arr, 1)
    else:
        driver = gdal.GetDriverByName("GTiff")
        ds = driver.Create(str(out_path), width, height, 1, gdal.GDT_Float32)
        # trf is a list: [top left x, w-e pixel size, rotation, top left y, rotation, n-s pixel size]
        ds.SetGeoTransform(trf_arr)
        ds.SetProjection(crs_wkt)
        ds.GetRasterBand(1).WriteArray(data_arr, 0, 0)
        ds.FlushCache()
        ds.SetNoDataValue(no_data_val)
        ds = None


def load_raster(
    path_str: str, bbox: list[int] | None = None, band: int = 0
) -> tuple[np.ndarray, list[float], str, int]:
    # Load raster, optionally crop to bbox
    path = check_path(path_str, make_dir=False)
    if not path.exists():
        raise FileNotFoundError(f"Raster file {path} does not exist.")
    if GDAL_ENV is False:
        with rasterio.open(path) as dataset:
            crs_wkt = dataset.crs.to_wkt() if dataset.crs is not None else None
            dataset_bounds = dataset.bounds
            no_data_val = dataset.nodata
            if bbox is not None:
                # Create bbox geometry for masking
                bbox_geom = geometry.box(*bbox)
                if not (
                    dataset_bounds.left <= bbox[0] <= dataset_bounds.right
                    and dataset_bounds.left <= bbox[2] <= dataset_bounds.right
                    and dataset_bounds.bottom <= bbox[1] <= dataset_bounds.top
                    and dataset_bounds.bottom <= bbox[3] <= dataset_bounds.top
                ):
                    raise ValueError("Bounding box is not fully contained within the raster dataset bounds")
                rast, trf = mask(dataset, [bbox_geom], crop=True)
            else:
                rast = dataset.read()
                trf = dataset.transform
            # Convert rasterio Affine to GDAL-style list
            trf_arr = [trf.c, trf.a, trf.b, trf.f, trf.d, trf.e]
            rast_arr = rast[band].astype(float)
    else:
        dataset = gdal.Open(str(path))
        if dataset is None:
            raise FileNotFoundError(f"Could not open {path}")
        trf = dataset.GetGeoTransform()
        crs_wkt = dataset.GetProjection().ExportToWkt()
        rast_arr = dataset.GetRasterBand(band + 1).ReadAsArray().astype(float)
        no_data_val = dataset.GetRasterBand(band + 1).GetNoDataValue()
        if bbox is not None:
            min_x, min_y, max_x, max_y = bbox
            xoff = int((min_x - trf[0]) / trf[1])
            yoff = int((trf[3] - max_y) / abs(trf[5]))
            xsize = int((max_x - min_x) / trf[1])
            ysize = int((max_y - min_y) / abs(trf[5]))
            rast_arr = rast_arr[yoff : yoff + ysize, xoff : xoff + xsize]
            trf_arr = [min_x, trf[1], 0, max_y, 0, trf[5]]
        else:
            trf_arr = [trf[0], trf[1], 0, trf[3], 0, trf[5]]
    if no_data_val is not None:
        rast_arr[rast_arr == no_data_val] = 0.0
    if rast_arr.min() < 0:
        raise ValueError("Raster contains negative values")
    return rast_arr, trf_arr, crs_wkt, no_data_val


def xy_to_lnglat(crs_wkt, x, y):
    """Convert x, y coordinates to longitude and latitude."""
    if GDAL_ENV is False:
        # Define the source and target CRS
        source_crs = pyproj.CRS(crs_wkt)
        target_crs = pyproj.CRS(4326)  # WGS 84
        # Create a transformer object
        transformer = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
        # Perform the transformation
        lng, lat = transformer.transform(x, y)
    else:
        # Define the source CRS from WKT
        old_cs = gdal.osr.SpatialReference()
        old_cs.ImportFromWkt(crs_wkt)
        # Define WGS 84 CRS
        new_cs = gdal.osr.SpatialReference()
        new_cs.ImportFromWkt("""
        GEOGCS["WGS 84",
            DATUM["WGS_1984",
                SPHEROID["WGS 84",6378137,298.257223563,
                    AUTHORITY["EPSG","7030"]],
                AUTHORITY["EPSG","6326"]],
            PRIMEM["Greenwich",0,
                AUTHORITY["EPSG","8901"]],
            UNIT["degree",0.01745329251994328,
                AUTHORITY["EPSG","9122"]],
            AUTHORITY["EPSG","4326"]]""")
        # Create a coordinate transformation
        transform = gdal.osr.CoordinateTransformation(old_cs, new_cs)
        lonlat = transform.TransformPoint(x, y)
        # Handle GDAL version differences
        gdalver = float(gdal.__version__[0])
        if gdalver >= 3.0:
            lng = lonlat[1]  # changed to gdal 3
            lat = lonlat[0]  # changed to gdal 3
        else:
            lng = lonlat[0]  # changed to gdal 2
            lat = lonlat[1]  # changed to gdal 2

    return lng, lat



# Generated with SMOP  0.41
from libsmop import *
# LinMeas_abs.m

    
@function
def LinMeas_abs(L=None,D=None,Lin=None,*args,**kwargs):
    varargin = LinMeas_abs.varargin
    nargin = LinMeas_abs.nargin

    #Used to Calculate the incoming longwave radiation absorbed by the cylinder with
#inputs epsilon (emissivity of the cylinder, 0.95), Ta (air temp, degreesC)
#L (length of cylinder, cm), D (diameter of cylinder, m)
    
    epsilon=0.95
# LinMeas_abs.m:7
    Acyl=CRT_Acyl(L,D)
# LinMeas_abs.m:9
    LinMeas_abs=multiply(multiply(multiply(epsilon,Lin),0.5),Acyl)
# LinMeas_abs.m:11
    
    #Written August, 2006 by Natasha Kenny
#Updated/Commented Dec 2015 Jenni Vanos


# %%
from pathlib import Path

from umep import (
    skyviewfactor_algorithm,
    wall_heightaspect_algorithm,
)
from umep.functions.SOLWEIGpython import solweig_runner_core

# %%
bbox = [476070, 4203550, 477110, 4204330]
working_folder = "temp/gothenburg"
pixel_resolution = 1  # metres
working_crs = 3007

working_path = Path(working_folder).absolute()
working_path.mkdir(parents=True, exist_ok=True)
working_path_str = str(working_path)

# input files for computing
dsm_path = "tests/data/Goteborg_SWEREF99_1200/DSM_KRbig.tif"
cdsm_path = "tests/data/Goteborg_SWEREF99_1200/CDSM_KRbig.tif"
config_path = "tests/data/Goteborg_SWEREF99_1200/configsolweig_alt.ini"
params_path = "tests/data/Goteborg_SWEREF99_1200/parametersforsolweig.json"

# %%
# setup parameters
trans_veg = 3
trunk_ratio = 0.25

# %%
# wall info for SOLWEIG (height and aspect)
if not Path.exists(working_path / "walls"):
    wall_heightaspect_algorithm.generate_wall_hts(
        dsm_path=dsm_path,
        bbox=None,
        out_dir=working_path_str + "/walls",
    )

# %%
# skyview factor for SOLWEIG
if not Path.exists(working_path / "svf"):
    skyviewfactor_algorithm.generate_svf(
        dsm_path=dsm_path,
        bbox=None,
        out_dir=working_path_str + "/svf",
        cdsm_path=cdsm_path,
        trans_veg=trans_veg,
        trunk_ratio=trunk_ratio,
    )

# %%
SWC = solweig_runner_core.SolweigRunCore(
    config_path_str=config_path,
    params_json_path=params_path,
)
SWC.run()
